from o3seespy.base_model import OpenSeesObject
import tempfile
import os


class RecorderBase(OpenSeesObject):
    op_base_type = "recorder"
    
    
class RecorderToArrayCacheBase(RecorderBase):  # TODO: implement NodeToArray where data saved to memory and loaded as array without collect
    fname = None
    
    def collect(self, unlink=True):
        from numpy import loadtxt
        try:
            a = loadtxt(self.fname, dtype=float)
        except ValueError as e:
            print('Warning: Need to run opy.wipe() before collecting arrays')
            raise ValueError(e)
        if unlink:
            try:
                os.unlink(self.fname)
            except PermissionError:
                print('Warning: Need to run opy.wipe() before collecting arrays')
        return a


class NodeToFile(RecorderBase):
    op_type = "Node"

    def __init__(self, osi, fname, node, dofs, res_type, nsd=8, dt=None, time=False):
        """
        Records properties of a node and saves the results to a file

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fname: str
            Full file name
        node: o3seespy.node.Node
        dofs: list
            A list of integers representing the degrees-of-freedom
        res_type: str
            Response type
        nsd: int
            Number of significant figures
        dt: float
            Time step
        time: bool
            If true the first column is the time
        """
        self.osi = osi
        self._parameters = [self.op_type, '-file', fname, '-precision', nsd, '-node', node.tag]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        if time:
            self._parameters.insert(5, '-time')
        self._parameters += ['-dof', *dofs, res_type]
        self._tag = self.to_process(osi)


class NodesToFile(RecorderBase):
    op_type = "Node"

    def __init__(self, osi, fname, nodes, dofs, res_type, nsd=8, dt=None, time=False):
        """
        Records properties of several nodes and saves the results to a file

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fname: str
            Full file name
        node: list
            List of o3seespy.node.Node objects
        dofs: list
            A list of integers representing the degrees-of-freedom
        res_type: str
            Response type
        nsd: int
            Number of significant figures
        dt: float
            Time step
        time: bool
            If true the first column is the time
        """
        self.osi = osi
        if isinstance(nodes, str) and nodes == 'all':
            node_tags = osi.to_process('getNodeTags', [])
        else:
            node_tags = [x.tag for x in nodes]
        self._parameters = [self.op_type, '-file', fname, '-precision', nsd, '-node', *node_tags, '-dof', *dofs, res_type]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        if time:
            self._parameters.insert(5, '-time')
        self._tag = self.to_process(osi)


class NodeToArrayCache(RecorderToArrayCacheBase):  # TODO: implement NodeToArray where data saved to memory and loaded as array without collect
    op_type = "Node"

    def __init__(self, osi, node, dofs, res_type, nsd=8, dt=None, fname=None):
        """
        Records properties of a node and saves results to a numpy array

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        node: o3seespy.node.Node
        dofs: list
            A list of integers representing the degrees-of-freedom
        res_type: str
            Response type: `disp`, 'vel', 'accel', 'incrDisp', 'reaction', 'eigin i', 'rayleighForces'
        nsd: int
            Number of significant figures
        dt: float
            Time step
        fname: str
            Full file path where data should be stored and loaded from
        """
        self.osi = osi
        if fname is None:
            self.fname = tempfile.NamedTemporaryFile(delete=False).name
        else:
            self.fname = fname
        self._parameters = [self.op_type, '-file', self.fname, '-precision', nsd, '-node', node.tag]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        self._parameters += ['-dof', *dofs, res_type]
        self._tag = self.to_process(osi)


class NodesToArrayCache(RecorderToArrayCacheBase):  # TODO: implement NodeToArray where data saved to memory and loaded as array without collect
    op_type = "Node"

    def __init__(self, osi, nodes, dofs, res_type, nsd=8, dt=None, fname=None):
        """
        Records properties of several nodes and saves results to a numpy array

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        nodes: list
            A list of o3seespy.node.Node objects
        dofs: list
           A list of integers representing the degrees-of-freedom
        res_type: str
           Response type
        nsd: int
           Number of significant figures
        dt: float
           Time step
        """
        self.osi = osi
        if isinstance(nodes, str) and nodes == 'all':
            node_tags = osi.to_process('getNodeTags', [])
        else:
            node_tags = [x.tag for x in nodes]
        if fname is None:
            self.fname = tempfile.NamedTemporaryFile(delete=False).name
        else:
            self.fname = fname
        self._parameters = [self.op_type, '-file', self.fname, '-precision', nsd, '-node', *node_tags, '-dof', *dofs, res_type]
        if dt is not None:
            self._parameters.insert(5,'-dT')
            self._parameters.insert(6, dt)
        self._tag = self.to_process(osi)


class TimeToArrayCache(RecorderBase):
    op_type = "Node"

    def __init__(self, osi, nsd=8, dt=None, fname=None):
        """
        Records the recorder time and saves to a numpy array

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        nsd: int
            Number of significant figures
        dt: float
            Time step
        """
        self.osi = osi
        if fname is None:
            self.fname = tempfile.NamedTemporaryFile(delete=False).name
        else:
            self.fname = fname
        self._parameters = [self.op_type, '-file', self.fname, '-precision', nsd, '-time', '-node', 1]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        self._parameters += ['-dof', 1, 'accel']
        self._tag = self.to_process(osi)

    def collect(self, unlink=True):
        from numpy import loadtxt
        try:
            a = loadtxt(self.fname, dtype=float)
        except ValueError as e:
            print('Warning: Need to run opy.wipe() before collecting arrays')
            raise ValueError(e)
        if unlink:
            try:
                os.unlink(self.fname)
            except PermissionError:
                print('Warning: Need to run opy.wipe() before collecting arrays')
        return a[:, 0]


class TimeToFile(RecorderBase):
    op_type = "Node"

    def __init__(self, osi, fname, nsd=8, dt=None):
        """
        Records the recorder time and saves to a numpy array

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        nsd: int
            Number of significant figures
        dt: float
            Time step
        """
        self.osi = osi
        self.fname = tempfile.NamedTemporaryFile(delete=False).name
        self._parameters = [self.op_type, '-file', fname, '-precision', nsd, '-time', '-node', 1]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        self._parameters += ['-dof', 1, 'disp']
        self._tag = self.to_process(osi)


class ElementToFile(RecorderBase):
    op_type = "Element"

    def __init__(self, osi, fname, ele, material=None, arg_vals=None, nsd=8, dt=None, time=False):
        """
        Records properties of an element and saves the results to a file

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fname: str
            Full file name
        ele: o3seespy.element.BaseElement
            An o3seespy element
        material: -
        arg_vals: list
            Extra arguments passed to element recorder method
        nsd: int
            Number of significant figures
        dt: float
            Time step
        time: bool
            If true the first column is the time
        """
        self.osi = osi
        if arg_vals is None:
            arg_vals = []
        extra_pms = []
        if material is not None:
            extra_pms += ['material', material]
        self._parameters = [self.op_type, '-file', fname, '-precision', nsd, '-ele', ele.tag, *extra_pms, *arg_vals]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        if time:
            self._parameters.insert(5, '-time')
        self._tag = self.to_process(osi)


class ElementsToFile(RecorderBase):
    op_type = "Element"

    def __init__(self, osi, fname, eles, material=None, arg_vals=None, nsd=8, dt=None, time=False):
        """
        Records properties of an element and saves the results to a file

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fname: str
            Full file name
        eles: list of o3seespy.element.BaseElement
            List of o3seespy elements
        material: -
        arg_vals: list
            Extra arguments passed to element recorder method
        nsd: int
            Number of significant figures
        dt: float
            Time step
        time: bool
            If true the first column is the time
        """
        self.osi = osi
        if arg_vals is None:
            arg_vals = []
        extra_pms = []
        if material is not None:
            extra_pms += ['material', material]
        self.ele_tags = [x.tag for x in eles]
        self._parameters = [self.op_type, '-file', fname, '-precision', nsd, '-ele', *self.ele_tags, *extra_pms, *arg_vals]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        if time:
            self._parameters.insert(5, '-time')
        self._tag = self.to_process(osi)


class ElementToArrayCache(RecorderToArrayCacheBase):  # TODO: implement ElementToArray where data saved to memory and loaded as array without collect
    op_type = "Element"

    def __init__(self, osi, ele, material=None, arg_vals=None, nsd=8, fname=None, dt=None):
        self.osi = osi
        if arg_vals is None:
            arg_vals = []
        self.arg_vals = [str(x) for x in arg_vals]
        extra_pms = []
        if material is not None:
            extra_pms += ['material', material]
        if fname is None:
            self.fname = tempfile.NamedTemporaryFile(delete=False).name
        else:
            self.fname = fname
        self.ele = ele
        self._parameters = [self.op_type, '-file', self.fname, '-precision', nsd, '-ele', ele.tag, *extra_pms, *self.arg_vals]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        self._tag = self.to_process(osi)


class ElementsToArrayCache(RecorderToArrayCacheBase):
    op_type = "Element"

    def __init__(self, osi, eles, material=None, arg_vals=None, nsd=8, fname=None, dt=None):
        self.osi = osi
        if arg_vals is None:
            arg_vals = []
        extra_pms = []
        if material is not None:
            extra_pms += ['material', material]
        if fname is None:
            self.fname = tempfile.NamedTemporaryFile(delete=False).name
        else:
            self.fname = fname
        self.ele_tags = [x.tag for x in eles]

        self._parameters = [self.op_type, '-file', self.fname, '-precision', nsd, '-ele', *self.ele_tags, *extra_pms, *arg_vals]
        if dt is not None:
            self._parameters.insert(5, '-dT')
            self._parameters.insert(6, dt)
        self._tag = self.to_process(osi)


def remove_recorder(osi, recorder):
    return osi.to_process('remove', ['recorder', recorder.tag])


def remove_recorders(osi):
    """Removes all recorders"""
    return osi.to_process('remove', ['recorders'])


def load_recorder_options():
    folder_path = os.path.dirname(os.path.realpath(__file__))
    return open(folder_path + '/mat_recorder_options.csv')



from o3seespy.base_model import OpenSeesObject


class GeomTransfBase(OpenSeesObject):
    op_base_type = "geomTransf"


class Linear2D(GeomTransfBase):
    """
    The Linear2D GeomTransf Class
    
    
    """
    op_type = 'Linear'

    def __init__(self, osi, d_i: list=None, d_j: list=None):
        """
        Initial method for Linear2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        d_i: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node i
            (the number of arguments depends on the dimensions of the current model).
        d_j: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node j
            (the number of arguments depends on the dimensions of the current model).

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> d_i = [1.0, 1.0]
        >>> d_j = [1.0, 1.0]
        >>> o3.geom_transf.Linear2D(osi, d_i=d_i, d_j=d_j)
        """
        self.osi = osi
        self.d_i = d_i
        self.d_j = d_j
        osi.n_transformation += 1
        self._tag = osi.n_transformation
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'd_i') is not None:
            self._parameters += ['-jntOffset', *self.d_i]
        if getattr(self, 'd_j') is not None:
            if getattr(self, 'd_i') is None:
                raise ValueError('Cannot set: d_j and not: d_i')
            self._parameters += [*self.d_j]
        self.to_process(osi)


class Linear3D(GeomTransfBase):
    """
    The Linear3D GeomTransf Class
    
    This command is used to construct a linear coordinate transformation (LinearCrdTransf) object, which performs a
    linear geometric transformation of beam stiffness and resisting force from the basic system to the
    global-coordinate system.
    """
    op_type = 'Linear'

    def __init__(self, osi, vecxz, d_i: list=None, d_j: list=None):
        """
        Initial method for Linear3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        vecxz: list
            X, y, and z components of vecxz, the vector used to define the local x-z plane of the local-coordinate
            system. the local y-axis is defined by taking the cross product of the vecxz vector and the x-axis. these
            components are specified in the global-coordinate system x,y,z and define a vector that is in a plane
            parallel to the x-z plane of the local-coordinate system. these items need to be specified for the
            three-dimensional problem.
        d_i: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node i
            (the number of arguments depends on the dimensions of the current model).
        d_j: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node j
            (the number of arguments depends on the dimensions of the current model).

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> vecxz = [1.0, 1.0]
        >>> d_i = [1.0, 1.0]
        >>> d_j = [1.0, 1.0]
        >>> o3.geom_transf.Linear3D(osi, vecxz=vecxz, d_i=d_i, d_j=d_j)
        """
        self.osi = osi
        self.vecxz = vecxz
        self.d_i = d_i
        self.d_j = d_j
        osi.n_transformation += 1
        self._tag = osi.n_transformation
        self._parameters = [self.op_type, self._tag, *self.vecxz]
        if getattr(self, 'd_i') is not None:
            self._parameters += ['-jntOffset', *self.d_i]
        if getattr(self, 'd_j') is not None:
            if getattr(self, 'd_i') is None:
                raise ValueError('Cannot set: d_j and not: d_i')
            self._parameters += [*self.d_j]
        self.to_process(osi)


class PDelta2D(GeomTransfBase):
    """
    The PDelta2D GeomTransf Class
    
    
    """
    op_type = 'PDelta'

    def __init__(self, osi, d_i: list=None, d_j: list=None):
        """
        Initial method for PDelta2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        d_i: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node i
            (the number of arguments depends on the dimensions of the current model).
        d_j: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node j
            (the number of arguments depends on the dimensions of the current model).

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> d_i = [1.0, 1.0]
        >>> d_j = [1.0, 1.0]
        >>> o3.geom_transf.PDelta2D(osi, d_i=d_i, d_j=d_j)
        """
        self.osi = osi
        self.d_i = d_i
        self.d_j = d_j
        osi.n_transformation += 1
        self._tag = osi.n_transformation
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'd_i') is not None:
            self._parameters += ['-jntOffset', *self.d_i]
        if getattr(self, 'd_j') is not None:
            if getattr(self, 'd_i') is None:
                raise ValueError('Cannot set: d_j and not: d_i')
            self._parameters += [*self.d_j]
        self.to_process(osi)


class PDelta3D(GeomTransfBase):
    """
    The PDelta3D GeomTransf Class
    
    This command is used to construct the P-Delta Coordinate Transformation (PDeltaCrdTransf) object, which performs a
    linear geometric transformation of beam stiffness and resisting force from the basic system to the global coordinate
    system, considering second-order P-Delta effects.
    """
    op_type = 'PDelta'

    def __init__(self, osi, vecxz, d_i: list=None, d_j: list=None):
        """
        Initial method for PDelta3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        vecxz: list
            X, y, and z components of vecxz, the vector used to define the local x-z plane of the local-coordinate
            system. the local y-axis is defined by taking the cross product of the vecxz vector and the x-axis. these
            components are specified in the global-coordinate system x,y,z and define a vector that is in a plane
            parallel to the x-z plane of the local-coordinate system. these items need to be specified for the
            three-dimensional problem.
        d_i: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node i
            (the number of arguments depends on the dimensions of the current model).
        d_j: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node j
            (the number of arguments depends on the dimensions of the current model).

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> vecxz = [1.0, 1.0]
        >>> d_i = [1.0, 1.0]
        >>> d_j = [1.0, 1.0]
        >>> o3.geom_transf.PDelta3D(osi, vecxz=vecxz, d_i=d_i, d_j=d_j)
        """
        self.osi = osi
        self.vecxz = vecxz
        self.d_i = d_i
        self.d_j = d_j
        osi.n_transformation += 1
        self._tag = osi.n_transformation
        self._parameters = [self.op_type, self._tag, *self.vecxz]
        if getattr(self, 'd_i') is not None:
            self._parameters += ['-jntOffset', *self.d_i]
        if getattr(self, 'd_j') is not None:
            if getattr(self, 'd_i') is None:
                raise ValueError('Cannot set: d_j and not: d_i')
            self._parameters += [*self.d_j]
        self.to_process(osi)


class Corotational2D(GeomTransfBase):
    """
    The Corotational2D GeomTransf Class
    
    
    """
    op_type = 'Corotational'

    def __init__(self, osi, d_i: list=None, d_j: list=None):
        """
        Initial method for Corotational2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        d_i: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node i
            (the number of arguments depends on the dimensions of the current model).
        d_j: list, optional
            Joint offset values -- offsets specified with respect to the global coordinate system for element-end node j
            (the number of arguments depends on the dimensions of the current model).

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> d_i = [1.0, 1.0]
        >>> d_j = [1.0, 1.0]
        >>> o3.geom_transf.Corotational2D(osi, d_i=d_i, d_j=d_j)
        """
        self.osi = osi
        self.d_i = d_i
        self.d_j = d_j
        osi.n_transformation += 1
        self._tag = osi.n_transformation
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'd_i') is not None:
            self._parameters += ['-jntOffset', *self.d_i]
        if getattr(self, 'd_j') is not None:
            if getattr(self, 'd_i') is None:
                raise ValueError('Cannot set: d_j and not: d_i')
            self._parameters += [*self.d_j]
        self.to_process(osi)


class Corotational3D(GeomTransfBase):
    """
    The Corotational3D GeomTransf Class
    
    This command is used to construct the Corotational Coordinate Transformation (CorotCrdTransf) object. Corotational
    transformation can be used in large displacement-small strain problems.
    """
    op_type = 'Corotational'

    def __init__(self, osi, vecxz):
        """
        Initial method for Corotational3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        vecxz: list
            X, y, and z components of vecxz, the vector used to define the local x-z plane of the local-coordinate
            system. the local y-axis is defined by taking the cross product of the vecxz vector and the x-axis. these
            components are specified in the global-coordinate system x,y,z and define a vector that is in a plane
            parallel to the x-z plane of the local-coordinate system. these items need to be specified for the
            three-dimensional problem.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> vecxz = [1.0, 1.0]
        >>> o3.geom_transf.Corotational3D(osi, vecxz=vecxz)
        """
        self.osi = osi
        self.vecxz = vecxz
        osi.n_transformation += 1
        self._tag = osi.n_transformation
        self._parameters = [self.op_type, self._tag, *self.vecxz]
        self.to_process(osi)



from o3seespy import opy

from tests.binary.functions import elastic_section, uniaxial_steel01_material, uniaxial_steel01_section


def test_define_uniaxial_steel101_section():
    opy.wipe()
    opy.model('basic', '-ndm', 2, '-ndf', 3)  # 2 dimensions, 3 dof per node
    section1_id = 1
    uniaxial_steel01_section(section1_id)


def test_beam_integration_mid_point_w_elastic():
    lp_i = 0.1
    lp_j = 0.2
    opy.wipe()
    opy.model('basic', '-ndm', 2, '-ndf', 3)  # 2 dimensions, 3 dof per node
    section_id = 1
    elastic_section(section_id)
    integ_tag = 1
    opy.beamIntegration('HingeMidpoint', integ_tag, section_id, lp_i, section_id, lp_j, section_id)


def test_beam_integration_mid_point_w_steel01():
    lp_i = 0.1
    lp_j = 0.2
    opy.wipe()
    opy.model('basic', '-ndm', 2, '-ndf', 3)  # 2 dimensions, 3 dof per node
    section1_id = 1
    section2_id = 2
    section3_id = 3
    uniaxial_steel01_section(section1_id)
    uniaxial_steel01_section(section2_id)
    uniaxial_steel01_section(section3_id)
    integ_tag = 1
    opy.beamIntegration('HingeMidpoint', integ_tag, section1_id, lp_i, section2_id, lp_j, section3_id)
    with_force_beam_column = 1
    if with_force_beam_column:
        # Establish nodes
        bot_node = 1
        top_node = 2
        transf_tag = 1
        opy.geomTransf('Linear', transf_tag, *[])
        opy.node(bot_node, 0., 0.)
        opy.node(top_node, 0., 5.)
        ele_i = 1
        opy.element('forceBeamColumn', ele_i, bot_node, top_node, transf_tag, integ_tag)


def test_beam_integration_lobatto_w_steel01():
    opy.wipe()
    opy.model('basic', '-ndm', 2, '-ndf', 3)  # 2 dimensions, 3 dof per node
    section_id = 1
    uniaxial_steel01_material(section_id)
    int_tag = 1
    opy.beamIntegration('Lobatto', int_tag, section_id, 8)
    assert 1 == 1  # This pass


if __name__ == '__main__':
    test_beam_integration_lobatto_w_steel01()
    test_beam_integration_mid_point_w_steel01()



import numpy as np
import o3seespy.extensions
import o3seespy as o3
import pytest

def test_2d_site_period():

    osi = o3.OpenSeesInstance(ndm=2, ndf=2, state=3)
    # Establish nodes
    node_depths = np.arange(0, 10, 1)
    n_node_rows = len(node_depths)
    x_nodes = np.arange(0, 10, 1)
    nx = len(x_nodes)
    nd = {}

    for yy in range(0, len(node_depths)):
        for xx in range(nx):
            # Establish left and right nodes
            nd[f"X{xx}Y{yy}"] = o3.node.Node(osi, x_nodes[xx], -node_depths[yy])
            # set x and y dofs equal for left and right nodes
        o3.EqualDOF(osi, nd[f"X0Y{yy}"], nd[f"X{nx - 1}Y{yy}"], [o3.cc.X])

    # Fix base nodes
    for xx in range(nx):
        o3.Fix2DOF(osi, nd[f"X{xx}Y{n_node_rows -1}"], o3.cc.FIXED, o3.cc.FIXED)
    for yy in range(0, len(node_depths) - 1):
        for xx in range(nx):
            o3.Fix2DOF(osi, nd[f"X{xx}Y{yy}"], o3.cc.FREE, o3.cc.FIXED)

    vs = 150.0
    rho = 1.8
    g_mod = vs ** 2 * rho
    poissons_ratio = 0.3
    e_mod = 2 * g_mod * (1 + poissons_ratio)
    ele_thick = 1.0

    soil_mat = o3.nd_material.ElasticIsotropic(osi, e_mod=e_mod, nu=poissons_ratio, rho=rho)
    eles = []
    for yy in range(0, len(node_depths) - 1):
        for xx in range(nx - 1):
            # def element
            nodes = [nd[f"X{xx}Y{yy + 1}"], nd[f"X{xx + 1}Y{yy + 1}"], nd[f"X{xx + 1}Y{yy}"], nd[f"X{xx}Y{yy}"]]
            eles.append(o3.element.SSPquad(osi, nodes, soil_mat, o3.cc.PLANE_STRAIN, ele_thick, 0.0, 0.0))

    # set damping based on first eigen mode
    angular_freq_sqrd = o3.get_eigen(osi, solver='fullGenLapack', n=1)
    if hasattr(angular_freq_sqrd, '__len__'):
        angular_freq = angular_freq_sqrd[0] ** 0.5
    else:
        angular_freq = angular_freq_sqrd ** 0.5
    # o3.extensions.to_py_file(osi, 'many_eles_2d.py')
    response_period = 2 * np.pi / angular_freq
    expected_period = 4 * max(node_depths) / vs
    print('response_period: ', response_period, expected_period, response_period / expected_period)
    assert np.isclose(response_period, expected_period, rtol=0.01)


def _run_dyn_1d_site_response(region_based=None):

    osi = o3.OpenSeesInstance(ndm=2, ndf=2, state=3)
    # Establish nodes
    node_depths = np.arange(0, 5, 1)
    n_node_rows = len(node_depths)
    x_nodes = np.arange(0, 2, 1)
    nx = len(x_nodes)
    nd = {}

    for yy in range(0, len(node_depths)):
        for xx in range(nx):
            # Establish left and right nodes
            nd[f"X{xx}Y{yy}"] = o3.node.Node(osi, x_nodes[xx], -node_depths[yy])
            # set x and y dofs equal for left and right nodes
        o3.EqualDOF(osi, nd[f"X0Y{yy}"], nd[f"X{nx - 1}Y{yy}"], [o3.cc.X])

    # Fix base nodes
    for xx in range(nx):
        o3.Fix2DOF(osi, nd[f"X{xx}Y{n_node_rows -1}"], o3.cc.FIXED, o3.cc.FIXED)
    for yy in range(0, len(node_depths) - 1):
        for xx in range(nx):
            o3.Fix2DOF(osi, nd[f"X{xx}Y{yy}"], o3.cc.FREE, o3.cc.FIXED)

    vs = 150.0
    rho = 1.8
    g_mod = vs ** 2 * rho
    poissons_ratio = 0.3
    e_mod = 2 * g_mod * (1 + poissons_ratio)
    ele_thick = 1.0

    soil_mat = o3.nd_material.ElasticIsotropic(osi, e_mod=e_mod, nu=poissons_ratio, rho=rho)
    eles = []
    for yy in range(0, len(node_depths) - 1):
        for xx in range(nx - 1):
            # def element
            nodes = [nd[f"X{xx}Y{yy + 1}"], nd[f"X{xx + 1}Y{yy + 1}"], nd[f"X{xx + 1}Y{yy}"], nd[f"X{xx}Y{yy}"]]
            eles.append(o3.element.SSPquad(osi, nodes, soil_mat, o3.cc.PLANE_STRAIN, ele_thick, 0.0, 0.0))

    freqs = np.array([0.5, 15.0])
    xi = 0.1  # high damping to see effects
    ang_f = np.pi / freqs
    alpha_m = xi * 2.0 * ang_f[0] * ang_f[1] / (ang_f[0] + ang_f[1])  # mass proportional
    beta_k = xi * 2.0 / (ang_f[0] + ang_f[1])  # stiffness proportional
    if region_based == 'node':
        o3.region.NodeRegion(osi, nodes='all', rayleigh={'alpha_m': alpha_m, 'beta_k_init': beta_k})
    if region_based == 'ele':
        o3.region.ElementRegion(osi, eles='all', rayleigh={'alpha_m': alpha_m, 'beta_k_init': beta_k})
    else:
        o3.rayleigh.Rayleigh(osi, alpha_m=alpha_m, beta_k=0.0, beta_k_init=beta_k, beta_k_comm=0.0)

    # Define the dynamic analysis
    dt = 0.01
    vals = np.sin(np.linspace(0, np.pi, 50))
    acc_series = o3.time_series.Path(osi, dt=dt, values=-vals)  # should be negative
    o3.pattern.UniformExcitation(osi, dir=o3.cc.X, accel_series=acc_series)
    # Run the dynamic analysis
    o3.wipe_analysis(osi)

    o3.algorithm.Newton(osi)
    o3.system.SparseGeneral(osi)
    o3.numberer.RCM(osi)
    o3.constraints.Transformation(osi)
    o3.integrator.Newmark(osi, 0.5, 0.25)
    o3.analysis.Transient(osi)

    o3.test_check.EnergyIncr(osi, tol=1.0e-3, max_iter=10)
    analysis_time = 1.0
    analysis_dt = 0.001
    rec_dt = 0.01
    tn = o3.recorder.NodeToArrayCache(osi, nd["X0Y0"], [o3.cc.DOF2D_X], 'accel', dt=rec_dt)

    if region_based:
        o3.extensions.to_py_file(osi, 'wr.py')
    else:
        o3.extensions.to_py_file(osi, 'wor.py')
    curr_time = o3.get_time(osi)
    while curr_time < analysis_time:
        status = o3.analyze(osi, 1, analysis_dt)
        curr_time = o3.get_time(osi)
        if status != 0:
            print('Not ok')
            print(status)

    o3.wipe(osi)

    x_acc = tn.collect()
    return x_acc


def test_region_based_damping():
    wnr = _run_dyn_1d_site_response(region_based='node')
    wer = _run_dyn_1d_site_response(region_based='ele')
    wor = _run_dyn_1d_site_response(region_based=None)
    assert np.isclose(wor, wnr).all()
    assert np.isclose(wor, wer).all()


if __name__ == '__main__':
    test_2d_site_period()
    test_region_based_damping()


import o3seespy as o3  # for testing only
import pytest


def test_elastic():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)


def test_elastic_pp():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ElasticPP(osi, e_mod=1.0, epsy_p=1.0, epsy_n=None, eps0=0.0)


def test_elastic_pp_gap():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ElasticPPGap(osi, e_mod=1.0, fy=1.0, gap=1.0, eta=0.0, damage='noDamage')


def test_ent():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ENT(osi, e_mod=1.0)



def test_parallel():
    osi = o3.OpenSeesInstance(ndm=2)
    mats = [o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None),
            o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)]
    factor_args = [1.0, 1.0]
    o3.uniaxial_material.Parallel(osi, mats=mats, factor_args=factor_args)



def test_series():
    osi = o3.OpenSeesInstance(ndm=2)
    mats = [o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None),
            o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)]
    o3.uniaxial_material.Series(osi, mats=mats)




from o3seespy.command.nd_material.base_material import NDMaterialBase


class InitialStateAnalysisWrapper(NDMaterialBase):
    """
    The InitialStateAnalysisWrapper NDMaterial Class
    
    The InitialStateAnalysisWrapper nDMaterial allows for the use of the InitialStateAnalysis command for setting
    initial conditions. The InitialStateAnalysisWrapper can be used with any nDMaterial. This material wrapper
    allows for the development of an initial stress field while maintaining the original geometry of the
    problem. An example analysis is provided below to demonstrate the use of this material wrapper object.
    """
    op_type = 'InitialStateAnalysisWrapper'

    def __init__(self, osi, n_d_mat, n_dim):
        """
        Initial method for InitialStateAnalysisWrapper

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        n_d_mat: obj
            The object of the associated ndmaterial object
        n_dim: int
            Number of dimensions (2 for 2d, 3 for 3d)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
        >>> o3.nd_material.InitialStateAnalysisWrapper(osi, n_d_mat=mat, n_dim=1)
        """
        self.osi = osi
        self.n_d_mat = n_d_mat
        self.n_dim = int(n_dim)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.n_d_mat.tag, self.n_dim]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_update_material_stage(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)

    def set_g_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'shearModulus', value, ele, eles)

    def set_bulk_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'bulkModulus', value, ele, eles)

    def set_friction_angle(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'frictionAngle', value, ele, eles)

    def set_cohesion(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'cohesion', value, ele, eles)


class InitStressNDMaterial(NDMaterialBase):
    """
    The InitStressNDMaterial NDMaterial Class
    
    This command is used to construct an Initial Stress material object.The stress-strain behaviour for this material is
    defined by another material.Initial Stress Material enables definition of initial stress for the material under
    consideration.The strain that corresponds to the initial stress will be calculated from the other material.
    """
    op_type = 'InitStressNDMaterial'

    def __init__(self, osi, other, init_stress, n_dim):
        """
        Initial method for InitStressNDMaterial

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        other: obj
            Object of the other material
        init_stress: float
            Initial stress
        n_dim: int
            Number of dimensions (e.g. if plane strain ndim=2)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
        >>> o3.nd_material.InitStressNDMaterial(osi, other=mat, init_stress=1.0, n_dim=2)
        """
        self.osi = osi
        self.other = other
        self.init_stress = float(init_stress)
        self.n_dim = int(n_dim)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.other.tag, self.init_stress, self.n_dim]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class InitStrainNDMaterial(NDMaterialBase):
    """
    The InitStrainNDMaterial NDMaterial Class
    
    This command is used to construct an Initial Strain material object. The stress-strain behaviour for this material
    is defined by another material. Initial Strain Material enables definition of initial strains for the material under
    consideration. The stress that corresponds to the initial strain will be calculated from the other material.
    """
    op_type = 'InitStrainNDMaterial'

    def __init__(self, osi, other, init_strain, n_dim):
        """
        Initial method for InitStrainNDMaterial

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        other: obj
            Object of the other material
        init_strain: float
            Initial strain
        n_dim: float
            Number of dimensions
        """
        self.osi = osi
        self.other = other
        self.init_strain = float(init_strain)
        self.n_dim = float(n_dim)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.other.tag, self.init_strain, self.n_dim]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)



import o3seespy as o3  # for testing only


def test_quad():
    osi = o3.OpenSeesInstance(ndm=2)
    crds_i = [1.0, 1.0]
    crds_j = [1.0, 1.0]
    crds_k = [1.0, 1.0]
    crds_l = [1.0, 1.0]
    conc_conf = o3.uniaxial_material.Concrete01(osi, fpc=-5.0, epsc0=-0.005, fpcu=-3.5, eps_u=-0.02)
    o3.patch.Quad(osi, conc_conf, num_subdiv_ij=1, num_subdiv_jk=1, crds_i=crds_i, crds_j=crds_j, crds_k=crds_k, crds_l=crds_l)


def test_rect():
    osi = o3.OpenSeesInstance(ndm=2)
    crds_i = [1.0, 1.0]
    crds_j = [1.0, 1.0]
    conc_conf = o3.uniaxial_material.Concrete01(osi, fpc=-5.0, epsc0=-0.005, fpcu=-3.5, eps_u=-0.02)
    o3.patch.Rect(osi, conc_conf, num_subdiv_y=1, num_subdiv_z=1, crds_i=crds_i, crds_j=crds_j)


def test_circ():
    osi = o3.OpenSeesInstance(ndm=2)
    center = [1.0, 1.0]
    rad = [1.0, 1.0]
    ang = [1.0, 1.0]
    conc_conf = o3.uniaxial_material.Concrete01(osi, fpc=-5.0, epsc0=-0.005, fpcu=-3.5, eps_u=-0.02)
    o3.patch.Circ(osi, conc_conf, num_subdiv_circ=1, num_subdiv_rad=1, center=center, rad=rad, ang=ang)




from o3seespy.command.element.base_element import ElementBase


class SimpleContact2D(ElementBase):
    """
    The SimpleContact2D Element Class
    
    This command is used to construct a SimpleContact2D element object.

    
    """
    op_type = 'SimpleContact2D'

    def __init__(self, osi, i_node, j_node, c_node, l_node, mat, g_tol, f_tol):
        """
        Initial method for SimpleContact2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        i_node: obj
            Retained nodes (-ndm 2 -ndf 2)
        j_node: obj
            Retained nodes (-ndm 2 -ndf 2)
        c_node: obj
            Constrained node (-ndm 2 -ndf 2)
        l_node: obj
            Lagrange multiplier node (-ndm 2 -ndf 2)
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        g_tol: float
            Gap tolerance
        f_tol: float
            Force tolerance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0)
        >>> c_node = o3.node.Node(osi, 0.0, 0.0)
        >>> l_node = o3.node.Node(osi, 0.0, 1.0)
        >>> mat = o3.nd_material.ContactMaterial2D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
        >>> o3.element.SimpleContact2D(osi, i_node=i_node, j_node=j_node, c_node=c_node, l_node=l_node, mat=mat,
        >>>                            g_tol=1.0, f_tol=1.0)
        """
        self.osi = osi
        self.i_node = i_node
        self.j_node = j_node
        self.c_node = c_node
        self.l_node = l_node
        self.mat = mat
        self.g_tol = float(g_tol)
        self.f_tol = float(f_tol)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.i_node.tag, self.j_node.tag, self.c_node.tag, self.l_node.tag, self.mat.tag, self.g_tol, self.f_tol]
        self.to_process(osi)


class SimpleContact3D(ElementBase):
    """
    The SimpleContact3D Element Class
    
    This command is used to construct a SimpleContact3D element object.

    
    """
    op_type = 'SimpleContact3D'

    def __init__(self, osi, i_node, j_node, k_node, l_node, c_node, lagr_node, mat, g_tol, f_tol):
        """
        Initial method for SimpleContact3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        i_node: obj
            Master nodes (-ndm 3 -ndf 3)
        j_node: obj
            Master nodes (-ndm 3 -ndf 3)
        k_node: obj
            Master nodes (-ndm 3 -ndf 3)
        l_node: obj
            Master nodes (-ndm 3 -ndf 3)
        c_node: obj
            Constrained node (-ndm 3 -ndf 3)
        lagr_node: obj
            Lagrange multiplier node (-ndm 3 -ndf 3)
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        g_tol: float
            Gap tolerance
        f_tol: float
            Force tolerance

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=3)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> k_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
        >>> c_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
        >>> l_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> lagr_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> mat = o3.nd_material.ContactMaterial3D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
        >>> o3.element.SimpleContact3D(osi, i_node=i_node, j_node=j_node, k_node=k_node, l_node=l_node, c_node=c_node,
        >>>                            lagr_node=lagr_node, mat=mat, g_tol=1.0, f_tol=1.0)
        """
        self.osi = osi
        self.i_node = i_node
        self.j_node = j_node
        self.k_node = k_node
        self.l_node = l_node
        self.c_node = c_node
        self.lagr_node = lagr_node
        self.mat = mat
        self.g_tol = float(g_tol)
        self.f_tol = float(f_tol)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.i_node.tag, self.j_node.tag, self.k_node.tag, self.l_node.tag, self.c_node.tag, self.lagr_node.tag, self.mat.tag, self.g_tol, self.f_tol]
        self.to_process(osi)


class BeamContact2D(ElementBase):
    """
    The BeamContact2D Element Class
    
    This command is used to construct a BeamContact2D element object.

    
    """
    op_type = 'BeamContact2D'

    def __init__(self, osi, i_node, j_node, s_node, l_node, mat, width, g_tol, f_tol, c_flag):
        """
        Initial method for BeamContact2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        i_node: obj
            Master nodes (-ndm 2 -ndf 3)
        j_node: obj
            Master nodes (-ndm 2 -ndf 3)
        s_node: obj
            Slave node (-ndm 2 -ndf 2)
        l_node: obj
            Lagrange multiplier node (-ndm 2 -ndf 2)
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        width: float
            The width of the wall represented by the beam element in plane strain
        g_tol: float
            Gap tolerance
        f_tol: float
            Force  tolerance
        c_flag: int
            Optional initial contact flag

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0)
        >>> s_node = o3.node.Node(osi, 0.0, 1.0)
        >>> l_node = o3.node.Node(osi, 0.0, 1.0)
        >>> mat = o3.nd_material.ContactMaterial2D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
        >>> o3.element.BeamContact2D(osi, i_node=i_node, j_node=j_node, s_node=s_node, l_node=l_node, mat=mat, width=1.0,
        >>>                          g_tol=1.0, f_tol=1.0, c_flag=1)
        """
        self.osi = osi
        self.i_node = i_node
        self.j_node = j_node
        self.s_node = s_node
        self.l_node = l_node
        self.mat = mat
        self.width = float(width)
        self.g_tol = float(g_tol)
        self.f_tol = float(f_tol)
        self.c_flag = int(c_flag)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.i_node.tag, self.j_node.tag, self.s_node.tag, self.l_node.tag, self.mat.tag, self.width, self.g_tol, self.f_tol, self.c_flag]
        self.to_process(osi)


class BeamContact3D(ElementBase):
    """
    The BeamContact3D Element Class
    
    This command is used to construct a BeamContact3D element object.

    
    """
    op_type = 'BeamContact3D'

    def __init__(self, osi, i_node, j_node, c_node, l_node, radius, crd_transf, mat, g_tol, f_tol, c_flag):
        """
        Initial method for BeamContact3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        i_node: obj
            Master nodes (-ndm 3 -ndf 6)
        j_node: obj
            Master nodes (-ndm 3 -ndf 6)
        c_node: obj
            Constrained node (-ndm 3 -ndf 3)
        l_node: obj
            Lagrange multiplier node (-ndm 3 -ndf 3)
        radius: float
            Constant radius of circular beam associated with beam element
        crd_transf: obj
            Unique integer object associated with previously-defined geometrictransf object
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        g_tol: float
            Gap tolerance
        f_tol: float
            Force tolerance
        c_flag: int
            Optional initial contact flag

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> osi.reset_model_params(ndm=3, ndf=3)
        >>> c_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> l_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> osi.reset_model_params(ndm=3, ndf=6)
        >>> mat = o3.nd_material.ContactMaterial3D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
        >>> crd_transf = o3.geom_transf.Linear3D(osi, vecxz=[1.0, 1.0], d_i=[1.0, 1.0], d_j=[1.0, 1.0])
        >>> o3.element.BeamContact3D(osi, i_node=i_node, j_node=j_node, c_node=c_node, l_node=l_node, radius=1.0,
        >>>                          crd_transf=crd_transf, mat=mat, g_tol=1.0, f_tol=1.0, c_flag=1)
        """
        self.osi = osi
        self.i_node = i_node
        self.j_node = j_node
        self.c_node = c_node
        self.l_node = l_node
        self.radius = float(radius)
        self.crd_transf = crd_transf
        self.mat = mat
        self.g_tol = float(g_tol)
        self.f_tol = float(f_tol)
        self.c_flag = int(c_flag)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.i_node.tag, self.j_node.tag, self.c_node.tag, self.l_node.tag, self.radius, self.crd_transf.tag, self.mat.tag, self.g_tol, self.f_tol, self.c_flag]
        self.to_process(osi)


class BeamEndContact3D(ElementBase):
    """
    The BeamEndContact3D Element Class
    
    This command is used to construct a BeamEndContact3D element object.

    
    """
    op_type = 'BeamEndContact3D'

    def __init__(self, osi, i_node, j_node, c_node, l_node, radius, g_tol, f_tol, c_flag):
        """
        Initial method for BeamEndContact3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        i_node: obj
            Master node from the beam (-ndm 3 -ndf 6)
        j_node: obj
            The remaining node on the beam element with ``inode`` (-ndm 3 -ndf 6)
        c_node: obj
            Constrained node (-ndm 3 -ndf 3)
        l_node: obj
            Lagrange multiplier node (-ndm 3 -ndf 3)
        radius: float
            Radius of circular beam associated with beam element
        g_tol: float
            Gap tolerance
        f_tol: float
            Force tolerance
        c_flag: float
            Optional initial contact flag

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> osi.reset_model_params(ndm=3, ndf=3)
        >>> c_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> l_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
        >>> o3.element.BeamEndContact3D(osi, i_node=i_node, j_node=j_node, c_node=c_node, l_node=l_node, radius=1.0,
        >>>                             g_tol=1.0, f_tol=1.0, c_flag=1.0)
        """
        self.osi = osi
        self.i_node = i_node
        self.j_node = j_node
        self.c_node = c_node
        self.l_node = l_node
        self.radius = float(radius)
        self.g_tol = float(g_tol)
        self.f_tol = float(f_tol)
        self.c_flag = float(c_flag)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.i_node.tag, self.j_node.tag, self.c_node.tag, self.l_node.tag, self.radius, self.g_tol, self.f_tol, self.c_flag]
        self.to_process(osi)



from o3seespy.command.element.base_element import ElementBase


class QuadUP(ElementBase):
    """
    The QuadUP Element Class
    
    FourNodeQuadUP is a four-node plane-strain element using bilinear isoparametric formulation. This element is
    implemented for simulating dynamic response of solid-fluid fully coupled material, based on Biot's theory of
    porous medium. Each element node has 3 degrees-of-freedom (DOF): DOF 1 and 2 for solid displacement (u) and
    DOF 3 for fluid pressure (p).

    
    """
    op_type = 'quadUP'

    def __init__(self, osi, ele_nodes, thick, mat, bulk, fmass, h_perm, v_perm, b1=0, b2=0, t=0):
        r"""
        Initial method for QuadUP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        thick: float
            Element thickness
        mat: obj
            Object of an ndmaterial object (previously defined) of which the element is composed
        bulk: float
            Combined undrained bulk modulus bc relating changes in pore pressure and volumetric strain, may be
            approximated by: :math:`b_c \approx b_f/n` where :math:`b_f` is the bulk modulus of fluid phase
            (:math:`2.2\times 10^6` kpa (or :math:`3.191\times 10^5` psi) for water), and n the initial porosity.
        fmass: float
            Fluid mass density
        h_perm: float
            Permeability coefficient in horizontal and vertical directions respectively.
        v_perm: float
            Permeability coefficient in horizontal and vertical directions respectively.
        b1: float, optional
            Optional gravity acceleration components in horizontal and vertical directions respectively (defaults are
            0.0)
        b2: float, optional
            Optional gravity acceleration components in horizontal and vertical directions respectively (defaults are
            0.0)
        t: float, optional
            Optional uniform element normal traction, positive in tension (default is 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.QuadUP(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat, bulk=1.0, fmass=1.0, h_perm=1.0, v_perm=1.0, b1=0, b2=0, t=0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.mat = mat
        self.bulk = float(bulk)
        self.fmass = float(fmass)
        self.h_perm = float(h_perm)
        self.v_perm = float(v_perm)
        self.b1 = float(b1)
        self.b2 = float(b2)
        self.t = float(t)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.mat.tag, self.bulk, self.fmass, self.h_perm, self.v_perm, self.b1, self.b2, self.t]
        self.to_process(osi)


class BrickUP(ElementBase):
    """
    The BrickUP Element Class
    
    BrickUP is an 8-node hexahedral linear isoparametric element. Each node has 4 degrees-of-freedom (DOF): DOFs 1 to 3
    for solid displacement (u) and DOF 4 for fluid pressure (p). This element is implemented for simulating dynamic
    response of solid-fluid fully coupled material, based on Biot's theory of porous medium.

    
    """
    op_type = 'brickUP'

    def __init__(self, osi, ele_nodes, mat, bulk, fmass, perm_x, perm_y, perm_z, b_x=0, b_y=0, b_z=0):
        r"""
        Initial method for BrickUP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of eight element nodes
        mat: obj
            Object of an ndmaterial object (previously defined) of which the element is composed
        bulk: float
            Combined undrained bulk modulus bc relating changes in pore pressure and volumetric strain, may be
            approximated by: :math:`b_c \approx b_f/n` where :math:`b_f` is the bulk modulus of fluid phase
            (:math:`2.2\times 10^6` kpa (or :math:`3.191\times 10^5` psi) for water), and n the initial porosity.
        fmass: float
            Fluid mass density
        perm_x: float
            Permeability coefficients in x, y, and z directions respectively.
        perm_y: float
            Permeability coefficients in x, y, and z directions respectively.
        perm_z: float
            Permeability coefficients in x, y, and z directions respectively.
        b_x: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)
        b_y: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)
        b_z: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.BrickUP(osi, ele_nodes=ele_nodes, mat=mat, bulk=1.0, fmass=1.0, perm_x=1.0, perm_y=1.0, perm_z=1.0, b_x=0, b_y=0, b_z=0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.bulk = float(bulk)
        self.fmass = float(fmass)
        self.perm_x = float(perm_x)
        self.perm_y = float(perm_y)
        self.perm_z = float(perm_z)
        self.b_x = float(b_x)
        self.b_y = float(b_y)
        self.b_z = float(b_z)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.bulk, self.fmass, self.perm_x, self.perm_y, self.perm_z, self.b_x, self.b_y, self.b_z]
        self.to_process(osi)


class BbarQuadUP(ElementBase):
    """
    The BbarQuadUP Element Class
    
    bbarQuadUP is a four-node plane-strain mixed volume/pressure element, which uses a tri-linear isoparametric
    formulation. This element is implemented for simulating dynamic response of solid-fluid fully coupled
    material, based on Biot's theory of porous medium. Each element node has 3 degrees-of-freedom (DOF):
    DOF 1 and 2 for solid displacement (u) and DOF 3 for fluid pressure (p).

    
    """
    op_type = 'bbarQuadUP'

    def __init__(self, osi, ele_nodes, thick, mat, bulk, fmass, h_perm, v_perm, b1=0, b2=0, t=0):
        r"""
        Initial method for BbarQuadUP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        thick: float
            Element thickness
        mat: obj
            Object of an ndmaterial object (previously defined) of which the element is composed
        bulk: float
            Combined undrained bulk modulus bc relating changes in pore pressure and volumetric strain, may be
            approximated by: :math:`b_c \approx b_f/n` where :math:`b_f` is the bulk modulus of fluid phase
            (:math:`2.2\times 10^6` kpa (or :math:`3.191\times 10^5` psi) for water), and n the initial porosity.
        fmass: float
            Fluid mass density
        h_perm: float
            Permeability coefficient in horizontal and vertical directions respectively.
        v_perm: float
            Permeability coefficient in horizontal and vertical directions respectively.
        b1: float, optional
            Optional gravity acceleration components in horizontal and vertical directions respectively (defaults are
            0.0)
        b2: float, optional
            Optional gravity acceleration components in horizontal and vertical directions respectively (defaults are
            0.0)
        t: float, optional
            Optional uniform element normal traction, positive in tension (default is 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.BbarQuadUP(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat, bulk=1.0, fmass=1.0, h_perm=1.0, v_perm=1.0, b1=0, b2=0, t=0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.mat = mat
        self.bulk = float(bulk)
        self.fmass = float(fmass)
        self.h_perm = float(h_perm)
        self.v_perm = float(v_perm)
        self.b1 = float(b1)
        self.b2 = float(b2)
        self.t = float(t)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.mat.tag, self.bulk, self.fmass, self.h_perm, self.v_perm, self.b1, self.b2, self.t]
        self.to_process(osi)


class BbarBrickUP(ElementBase):
    """
    The BbarBrickUP Element Class
    
    bbarBrickUP is a 8-node mixed volume/pressure element, which uses a tri-linear isoparametric formulation.Each node
    has 4 degrees-of-freedom (DOF): DOFs 1 to 3 for solid displacement (u) and DOF 4 for fluid pressure (p). This element
    is implemented for simulating dynamic response of solid-fluid fully coupled material, based on Biot's theory of
    porous medium.

    
    """
    op_type = 'bbarBrickUP'

    def __init__(self, osi, ele_nodes, mat, bulk, fmass, perm_x, perm_y, perm_z, b_x=0, b_y=0, b_z=0):
        r"""
        Initial method for BbarBrickUP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of eight element nodes
        mat: obj
            Object of an ndmaterial object (previously defined) of which the element is composed
        bulk: float
            Combined undrained bulk modulus bc relating changes in pore pressure and volumetric strain, may be
            approximated by: :math:`b_c \approx b_f/n` where :math:`b_f` is the bulk modulus of fluid phase
            (:math:`2.2\times 10^6` kpa (or :math:`3.191\times 10^5` psi) for water), and n the initial porosity.
        fmass: float
            Fluid mass density
        perm_x: float
            Permeability coefficients in x, y, and z directions respectively.
        perm_y: float
            Permeability coefficients in x, y, and z directions respectively.
        perm_z: float
            Permeability coefficients in x, y, and z directions respectively.
        b_x: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)
        b_y: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)
        b_z: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.BbarBrickUP(osi, ele_nodes=ele_nodes, mat=mat, bulk=1.0, fmass=1.0, perm_x=1.0, perm_y=1.0, perm_z=1.0, b_x=0, b_y=0, b_z=0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.bulk = float(bulk)
        self.fmass = float(fmass)
        self.perm_x = float(perm_x)
        self.perm_y = float(perm_y)
        self.perm_z = float(perm_z)
        self.b_x = float(b_x)
        self.b_y = float(b_y)
        self.b_z = float(b_z)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.bulk, self.fmass, self.perm_x, self.perm_y, self.perm_z, self.b_x, self.b_y, self.b_z]
        self.to_process(osi)


class N94QuadUP(ElementBase):
    """
    The N94QuadUP Element Class
    
    Nine_Four_Node_QuadUP is a 9-node quadrilateral plane-strain element. The four corner nodes have 3
    degrees-of-freedom (DOF) each: DOF 1 and 2 for solid displacement (u) and DOF 3 for fluid pressure
    (p). The other five nodes have 2 DOFs each for solid displacement. This element is implemented
    for simulating dynamic response of solid-fluid fully coupled material, based on Biot's theory of porous medium.

    
    """
    op_type = '9_4_QuadUP'

    def __init__(self, osi, ele_nodes, thick, mat, bulk, fmass, h_perm, v_perm, b1=0, b2=0):
        r"""
        Initial method for N94QuadUP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of nine element nodes
        thick: float
            Element thickness
        mat: obj
            Object of an ndmaterial object (previously defined) of which the element is composed
        bulk: float
            Combined undrained bulk modulus bc relating changes in pore pressure and volumetric strain, may be
            approximated by: :math:`b_c \approx b_f/n` where :math:`b_f` is the bulk modulus of fluid phase
            (:math:`2.2\times 10^6` kpa (or :math:`3.191\times 10^5` psi) for water), and n the initial porosity.
        fmass: float
            Fluid mass density
        h_perm: float
            Permeability coefficient in horizontal and vertical directions respectively.
        v_perm: float
            Permeability coefficient in horizontal and vertical directions respectively.
        b1: float, optional
            Optional gravity acceleration components in horizontal and vertical directions respectively (defaults are
            0.0)
        b2: float, optional
            Optional gravity acceleration components in horizontal and vertical directions respectively (defaults are
            0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1], [0.5, 0.5, 0.5]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.N94QuadUP(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat, bulk=1.0, fmass=1.0, h_perm=1.0, v_perm=1.0, b1=0, b2=0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.mat = mat
        self.bulk = float(bulk)
        self.fmass = float(fmass)
        self.h_perm = float(h_perm)
        self.v_perm = float(v_perm)
        self.b1 = float(b1)
        self.b2 = float(b2)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.mat.tag, self.bulk, self.fmass, self.h_perm, self.v_perm, self.b1, self.b2]
        self.to_process(osi)


class N208BrickUP(ElementBase):
    """
    The N208BrickUP Element Class
    
    Twenty_Eight_Node_BrickUP is a 20-node hexahedral isoparametric element.The eight corner nodes have 4
    degrees-of-freedom (DOF) each: DOFs 1 to 3 for solid displacement (u) and DOF 4 for fluid pressure (p).
    The other nodes have 3 DOFs each for solid displacement. This element is implemented for simulating
    dynamic response of solid-fluid fully coupled material, based on Biot's theory of porous medium.

    
    """
    op_type = '20_8_BrickUP'

    def __init__(self, osi, ele_nodes, mat, bulk, fmass, perm_x, perm_y, perm_z, b_x=0, b_y=0, b_z=0):
        r"""
        Initial method for N208BrickUP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of twenty element nodes
        mat: obj
            Object of an ndmaterial object (previously defined) of which the element is composed
        bulk: float
            Combined undrained bulk modulus bc relating changes in pore pressure and volumetric strain, may be
            approximated by: :math:`b_c \approx b_f/n` where :math:`b_f` is the bulk modulus of fluid phase
            (:math:`2.2\times 10^6` kpa (or :math:`3.191\times 10^5` psi) for water), and n the initial porosity.
        fmass: float
            Fluid mass density
        perm_x: float
            Permeability coefficients in x, y, and z directions respectively.
        perm_y: float
            Permeability coefficients in x, y, and z directions respectively.
        perm_z: float
            Permeability coefficients in x, y, and z directions respectively.
        b_x: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)
        b_y: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)
        b_z: float, optional
            Optional gravity acceleration components in x, y, and z directions directions respectively (defaults are
            0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.N208BrickUP(osi, ele_nodes=ele_nodes, mat=mat, bulk=1.0, fmass=1.0, perm_x=1.0, perm_y=1.0, perm_z=1.0, b_x=0, b_y=0, b_z=0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.bulk = float(bulk)
        self.fmass = float(fmass)
        self.perm_x = float(perm_x)
        self.perm_y = float(perm_y)
        self.perm_z = float(perm_z)
        self.b_x = float(b_x)
        self.b_y = float(b_y)
        self.b_z = float(b_z)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.bulk, self.fmass, self.perm_x, self.perm_y, self.perm_z, self.b_x, self.b_y, self.b_z]
        self.to_process(osi)



from o3seespy.command.uniaxial_material.base_material import UniaxialMaterialBase



class PySimple1(UniaxialMaterialBase):
    """
    The PySimple1 UniaxialMaterial Class
    
    This command is used to construct a PySimple1 uniaxial material object.
    """
    op_type = 'PySimple1'

    def __init__(self, osi, soil_type, pult, y50, cd, c=0.0):
        """
        Initial method for PySimple1

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        soil_type: int
            Soiltype = 1 backbone of p-y curve approximates matlock (1970) soft clay relation. soiltype = 2 backbone of
            p-y curve approximates api (1993) sand relation.
        pult: float
            Ultimate capacity of the p-y material. note that "p" or "pult" are distributed loads [force per length of
            pile] in common design equations, but are both loads for this uniaxial_material [i.e., distributed load times the
            tributary length of the pile].
        y50: float
            Displacement at which 50% of pult is mobilized in monotonic loading.
        cd: float
            Variable that sets the drag resistance within a fully-mobilized gap as cd*pult.
        c: float, optional
            The viscous damping term (dashpot) on the far-field (elastic) component of the displacement rate (velocity).
            (optional default = 0.0). nonzero c values are used to represent radiation damping effects

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.PySimple1(osi, soil_type=1, pult=1.0, y50=1.0, cd=1.0, c=0.0)
        """
        self.osi = osi
        self.soil_type = int(soil_type)
        self.pult = float(pult)
        self.y50 = float(y50)
        self.cd = float(cd)
        self.c = float(c)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.soil_type, self.pult, self.y50, self.cd, self.c]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class TzSimple1(UniaxialMaterialBase):
    """
    The TzSimple1 UniaxialMaterial Class
    
    This command is used to construct a TzSimple1 uniaxial material object.
    """
    op_type = 'TzSimple1'

    def __init__(self, osi, soil_type, tult, z50, c=0.0):
        """
        Initial method for TzSimple1

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        soil_type: int
            Soiltype = 1 backbone of t-z curve approximates reese and o'neill (1987). soiltype = 2 backbone of t-z curve
            approximates mosher (1984) relation.
        tult: float
            Ultimate capacity of the t-z material. see note 1.
        z50: float
            Displacement at which 50% of tult is mobilized in monotonic loading.
        c: float, optional
            The viscous damping term (dashpot) on the far-field (elastic) component of the displacement rate (velocity).
            (optional default = 0.0). see note 2.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.TzSimple1(osi, soil_type=1, tult=1.0, z50=1.0, c=0.0)
        """
        self.osi = osi
        self.soil_type = int(soil_type)
        self.tult = float(tult)
        self.z50 = float(z50)
        self.c = float(c)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.soil_type, self.tult, self.z50, self.c]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class QzSimple1(UniaxialMaterialBase):
    """
    The QzSimple1 UniaxialMaterial Class
    
    This command is used to construct a QzSimple1 uniaxial material object.
    """
    op_type = 'QzSimple1'

    def __init__(self, osi, qz_type, qult, z50, suction=0.0, c=0.0):
        """
        Initial method for QzSimple1

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        qz_type: int
            Qztype = 1 backbone of q-z curve approximates reese and o'neill's (1987) relation for drilled shafts in
            clay. qztype = 2 backbone of q-z curve approximates vijayvergiya's (1977) relation for piles in sand.
        qult: float
            Ultimate capacity of the q-z material. see note 1.
        z50: float
            Displacement at which 50% of qult is mobilized in monotonic loading. see note 2.
        suction: float, optional
            Uplift resistance is equal to suction*qult. default = 0.0. the value of suction must be 0.0 to 0.1.*
        c: float, optional
            The viscous damping term (dashpot) on the far-field (elastic) component of the displacement rate (velocity).
            default = 0.0. nonzero c values are used to represent radiation damping effects.*

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.QzSimple1(osi, qz_type=1, qult=1.0, z50=1.0, suction=0.0, c=0.0)
        """
        self.osi = osi
        self.qz_type = int(qz_type)
        self.qult = float(qult)
        self.z50 = float(z50)
        self.suction = float(suction)
        self.c = float(c)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.qz_type, self.qult, self.z50, self.suction, self.c]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class PyLiq1(UniaxialMaterialBase):
    """
    The PyLiq1 UniaxialMaterial Class
    
    
    """
    op_type = 'PyLiq1'

    def __init__(self, osi, soil_type, pult, y50, cd, c, p_res, ele1, ele2, time_series=None):
        """
        Initial method for PyLiq1

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        soil_type: int
            Soiltype = 1 backbone of p-y curve approximates matlock (1970) soft clay relation. soiltype = 2 backbone of
            p-y curve approximates api (1993) sand relation.
        pult: float
            Ultimate capacity of the p-y material. note that "p" or "pult" are distributed loads [force per length of
            pile] in common design equations, but are both loads for this uniaxial_material [i.e., distributed load times the
            tributary length of the pile].
        y50: float
            Displacement at which 50% of pult is mobilized in monotonic loading.
        cd: float
            Variable that sets the drag resistance within a fully-mobilized gap as cd*pult.
        c: float
            The viscous damping term (dashpot) on the far-field (elastic) component of the displacement rate (velocity).
            (optional default = 0.0). nonzero c values are used to represent radiation damping effects
        p_res: float
            Sets the minimum (or residual) peak resistance that the material retains as the adjacent solid soil elements
            liquefy
        ele1: float
            Are the eleobject (element numbers) for the two solid elements from which pyliq1 will obtain mean effective
            stresses and excess pore pressures
        ele2: float
            Are the eleobject (element numbers) for the two solid elements from which pyliq1 will obtain mean effective
            stresses and excess pore pressures
        time_series: obj, optional
            Alternatively, mean effective stress can be supplied by a time series by specifying the text string
            ``'-timeseries'`` and the object of the series    ``seriesobject``.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.PyLiq1(osi, soil_type=1, pult=1.0, y50=1.0, cd=1.0, c=1.0, p_res=1.0, ele1=1.0, ele2=1.0)
        """
        self.osi = osi
        self.soil_type = int(soil_type)
        self.pult = float(pult)
        self.y50 = float(y50)
        self.cd = float(cd)
        self.c = float(c)
        self.p_res = float(p_res)
        self.ele1 = float(ele1)
        self.ele2 = float(ele2)
        self.time_series = time_series
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.soil_type, self.pult, self.y50, self.cd, self.c, self.p_res, self.ele1, self.ele2]
        if getattr(self, 'time_series') is not None:
            self._parameters += ['-timeSeries', self.time_series.tag]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_update_material_stage(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)


class TzLiq1(UniaxialMaterialBase):
    """
    The TzLiq1 UniaxialMaterial Class
    
    
    """
    op_type = 'TzLiq1'

    def __init__(self, osi, tz_type, tult, z50, c, ele1, ele2, time_series=None):
        """
        Initial method for TzLiq1

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tz_type: int
            Tztype = 1 backbone of t-z curve approximates reese and o'neill (1987). tztype = 2 backbone of t-z curve
            approximates mosher (1984) relation.
        tult: float
            Ultimate capacity of the t-z material. see note 1.
        z50: float
            Displacement at which 50% of tult is mobilized in monotonic loading.
        c: float
            The viscous damping term (dashpot) on the far-field (elastic) component of the displacement rate (velocity).
        ele1: float
            Are the eleobject (element numbers) for the two solid elements from which pyliq1 will obtain mean effective
            stresses and excess pore pressures
        ele2: float
            Are the eleobject (element numbers) for the two solid elements from which pyliq1 will obtain mean effective
            stresses and excess pore pressures
        time_series: obj, optional
            Alternatively, mean effective stress can be supplied by a time series by specifying the text string
            ``'-timeseries'`` and the object of the seriesm    ``seriesobject``.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.TzLiq1(osi, tz_type=1, tult=1.0, z50=1.0, c=1.0, ele1=1.0, ele2=1.0)
        """
        self.osi = osi
        self.tz_type = int(tz_type)
        self.tult = float(tult)
        self.z50 = float(z50)
        self.c = float(c)
        self.ele1 = float(ele1)
        self.ele2 = float(ele2)
        self.time_series = time_series
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.tz_type, self.tult, self.z50, self.c, self.ele1, self.ele2]
        if getattr(self, 'time_series') is not None:
            self._parameters += ['-timeSeries', self.time_series.tag]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_update_material_stage(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)



import o3seespy as o3  # for testing only
import pytest


def test_elastic2d():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0, g_mod=0.0, alpha_y=0.0)


def test_elastic3d():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.section.Elastic3D(osi, e_mod=1.0, area=1.0, iz=1.0, iy=1.0, g_mod=1.0, jxx=1.0, alpha_y=0.0, alpha_z=0.0)


def test_fiber():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.uniaxial_material.ElasticPP(osi, e_mod=1.0, epsy_p=1.0, epsy_n=None, eps0=0.0)
    o3.section.Fiber(osi, torsion_mat=mat)


def test_fiber_thermal():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.section.FiberThermal(osi, gj=0.0)


def test_nd_fiber():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.section.NDFiber(osi)


def test_wf_section2d():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.uniaxial_material.Elastic(osi, 1.0)
    o3.section.WFSection2D(osi, mat=mat, d=1.0, tw=1.0, bf=1.0, tf=1.0, nfw=1.0, nff=1.0)


def test_rc_section2d():
    osi = o3.OpenSeesInstance(ndm=2)
    core_mat = o3.uniaxial_material.Concrete01(osi, -6.0, -0.004, -5.0, -0.014)
    cover_mat = o3.uniaxial_material.Concrete01(osi, -5.0, -0.002, 0.0, -0.006)
    steel_mat = o3.uniaxial_material.Steel01(osi, 60.0, 30000.0, 0.01)
    o3.section.RCSection2D(osi, core_mat=core_mat, cover_mat=cover_mat, steel_mat=steel_mat, d=1.0, b=1.0, cover_depth=1.0, atop=1.0, abot=1.0, aside=1.0, nfcore=1.0, nfcover=1.0, nfs=1.0)


def test_rc_circular_section():
    osi = o3.OpenSeesInstance(ndm=2)
    core_mat = o3.uniaxial_material.Concrete01(osi, -6.0, -0.004, -5.0, -0.014)
    cover_mat = o3.uniaxial_material.Concrete01(osi, -5.0, -0.002, 0.0, -0.006)
    steel_mat = o3.uniaxial_material.Steel01(osi, 60.0, 30000.0, 0.01)
    o3.section.RCCircularSection(osi, core_mat=core_mat, cover_mat=cover_mat, steel_mat=steel_mat, d=1.0,
                                 cover_depth=0.10, a_s=0.1, nrings_core=2, nrings_cover=2, newedges=2, nsteel=4, gj=0.0)


def test_parallel():
    osi = o3.OpenSeesInstance(ndm=2)
    secs = [o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0, g_mod=0.0, alpha_y=0.0),
             o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0, g_mod=0.0, alpha_y=0.0)]
    o3.section.Parallel(osi, secs)


def test_aggregator():
    osi = o3.OpenSeesInstance(ndm=2)
    section = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    mats = [[o3.uniaxial_material.Elastic(osi, 1.0, 1.0), o3.cc.P], [o3.uniaxial_material.Elastic(osi, 1.0, 1.0), o3.cc.M_Z]]
    o3.section.Aggregator(osi, mats=mats, section=section)


def test_uniaxial():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.uniaxial_material.Elastic(osi, 1.0)
    o3.section.Uniaxial(osi, mat=mat, quantity='P')


def test_elastic_membrane_plate_section():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.section.ElasticMembranePlateSection(osi, e_mod=1.0, nu=1.0, h=1.0, rho=1.0)


@pytest.mark.skip()  # needs update to opensees
def test_plate_fiber():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.uniaxial_material.Elastic(osi, 1.0)
    o3.section.PlateFiber(osi, mat=mat, h=1.0)


def test_bidirectional():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.section.Bidirectional(osi, e_mod=1.0, fy=1.0, hiso=1.0, hkin=1.0, code1='Vy', code2='P')


def test_iso2spring():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.section.Isolator2spring(osi, tol=1.0, k1=1.0, fyo=1.0, k2o=1.0, kvo=1.0, hb=1.0, pe=1.0, po=0.0)


@pytest.mark.skip()  # unexpected issue with opensees implementation
def test_layered_shell():
    osi = o3.OpenSeesInstance(ndm=2)
    mats = [[o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, v=1.0, rho=0.0), 1.0],
            [o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, v=1.0, rho=0.0), 1.0],
            [o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, v=1.0, rho=0.0), 1.0]]
    o3.section.LayeredShell(osi, mats=mats)




import o3seespy as o3  # for testing only
import pytest


def test_beam_column_joint():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    mats = [o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None) for x in range(13)]
    o3.element.BeamColumnJoint(osi, ele_nodes, *mats, ele_height_fac=1.0, ele_width_fac=1.0)


def test_elastic_tubular_joint():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.ElasticTubularJoint(osi, ele_nodes=ele_nodes, brace__diameter=1.0, brace__angle=1.0, big_e=1.0, chord__diameter=1.0, chord__thickness=1.0, chord__angle=1.0)


@pytest.mark.skip()
def test_joint2d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1], [0.5, 0.5]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.Joint2D(osi, ele_nodes=ele_nodes, mat1=1, mat2=1, mat3=1, mat4=1, mat_c=1, lrg_dsp='', dmg='', dmg1dmg2dmg3dmg4dmg_c=1)




import os
import o3seespy as o3


class Results2D(object):
    coords = None
    time = None
    x_disp = None
    y_disp = None
    node_c = None
    ele_c = None
    used_r_starter = 0
    mat2ele_tags = None  # Assume 1-to-1 so if it uses a section then should be null
    sect2ele_tags = None  # Store position and tag - UNUSED
    mat2sect_tags = None  # UNUSED  # TODO: implement
    n_nodes_per_ele = [2, 4, 8]  # for 2D

    def __init__(self, cache_path='', dt=None, dynamic=False):
        self.cache_path = cache_path
        self._dt = dt
        self.dynamic = dynamic
        from numpy import savetxt, loadtxt
        self.savetxt = savetxt
        self.loadtxt = loadtxt
        self.ele2node_tags = {}
        self.meta_files = ['node_c', 'ele_c', 'mat2ele_tags', 'sect2ele_tags', 'mat2sect_tags']
        self.meta_fmt = [None, '%i', '%i', '%i']
        self.pseudo_dt = None  # use if recording steps of a static analysis

    def start_recorders(self, osi, dt=None):  # TODO: handle recorder time step
        self.used_r_starter = 1
        if self.coords is None:
            self.coords = o3.get_all_node_coords(osi)
        if not self.ele2node_tags:
            self.ele2node_tags = o3.get_all_ele_node_tags_as_dict(osi)
        if dt is not None:
            self._dt = dt
        if self.dynamic:
            o3.recorder.NodesToFile(osi, f'{self.cache_path}x_disp.txt', 'all', [o3.cc.DOF2D_X], 'disp', nsd=4, dt=dt)
            o3.recorder.NodesToFile(osi, f'{self.cache_path}y_disp.txt', 'all', [o3.cc.DOF2D_Y], 'disp', nsd=4, dt=dt)
            if not self.pseudo_dt:
                o3.recorder.TimeToFile(osi, f'{self.cache_path}timer.txt', nsd=4, dt=dt)

    def wipe_old_files(self):
        try:
            os.remove(f'{self.cache_path}ele2node_tags.txt')
        except FileNotFoundError:
            pass
        try:
            os.remove(f'{self.cache_path}coords.txt')
        except FileNotFoundError:
            pass
        if not self.used_r_starter:
            try:
                os.remove(f'{self.cache_path}x_disp.txt')
            except FileNotFoundError:
                pass
            try:
                os.remove(f'{self.cache_path}y_disp.txt')
            except FileNotFoundError:
                pass

        for fname in self.meta_files:
            try:
                os.remove(f'{self.cache_path}{fname}.txt')
            except FileNotFoundError:
                pass

    def save_to_cache(self):
        self.wipe_old_files()
        self.savetxt(self.cache_path + 'coords.txt', self.coords)
        ostr = [f'{ele_tag} ' + ' '.join([str(x) for x in self.ele2node_tags[ele_tag]]) + '\n' for ele_tag in self.ele2node_tags]
        open(self.cache_path + 'ele2node_tags.txt', 'w').writelines(ostr)

        for i, fname in enumerate(self.meta_files):
            vals = getattr(self, fname)
            if vals is not None:
                self.savetxt(self.cache_path + f'{fname}.txt', vals, fmt=self.meta_fmt[i])
        if self.dynamic:
            if not self.used_r_starter:
                self.savetxt(self.cache_path + 'x_disp.txt', self.x_disp)
                self.savetxt(self.cache_path + 'y_disp.txt', self.y_disp)
                self.savetxt(self.cache_path + 'timer.txt', self.time)
            elif self.pseudo_dt:
                from numpy import arange
                x_disp = self.loadtxt(f'{self.cache_path}x_disp.txt', ndmin=2)
                self.time = arange(len(x_disp[:, 0])) * self.pseudo_dt
                self.savetxt(self.cache_path + 'timer.txt', self.time)

    def load_from_cache(self):
        self.coords = self.loadtxt(self.cache_path + 'coords.txt')
        self.ele2node_tags = {}
        lines = open(self.cache_path + 'ele2node_tags.txt').read().splitlines()
        for line in lines:
            parts = [int(x) for x in line.split()]
            self.ele2node_tags[parts[0]] = parts[1:]
        for fname in self.meta_files:
            try:
                data = self.loadtxt(self.cache_path + f'{fname}.txt')
                if len(data) == 0:
                    data = None
                setattr(self, fname, data)
            except OSError:
                pass

        if self.dynamic:
            self.x_disp = self.loadtxt(f'{self.cache_path}x_disp.txt')
            self.y_disp = self.loadtxt(f'{self.cache_path}y_disp.txt')
            self.time = self.loadtxt(f'{self.cache_path}timer.txt', ndmin=2)[:, 0]

    @property
    def dt(self):
        if self._dt is None:
            if self.time is not None:
                self._dt = (self.time[-1] - self.time[0]) / (len(self.time) - 1)
        return self._dt

    @dt.setter
    def dt(self, dt):
        self._dt = dt




import o3seespy as o3  # for testing only


def test_straight():
    osi = o3.OpenSeesInstance(ndm=2)
    start = [1.0, 1.0]
    end = [1.0, 1.0]
    rebar = o3.uniaxial_material.Steel01(osi, fy=60.0, e0=30000.0, b=0.02)
    o3.layer.Straight(osi, rebar, num_fiber=1, area_fiber=1.0, start=start, end=end)


def test_circ():
    osi = o3.OpenSeesInstance(ndm=2)
    center = [1.0, 1.0]
    rebar = o3.uniaxial_material.Steel01(osi, fy=60.0, e0=30000.0, b=0.02)
    o3.layer.Circ(osi, rebar, num_fiber=6, area_fiber=1.0, center=center, radius=1.0, ang=[0.0, 360.0-360/6])



import o3seespy as o3  # for testing only


def test_cyc_liq_cp():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.CycLiqCP(osi, g0=1.0, kappa=1.0, h=1.0, mfc=1.0, dre1=1.0, mdc=1.0, dre2=1.0, rdr=1.0, alpha=1.0, dir=1.0, ein=1.0, rho=1.0)


def test_cyc_liq_cpsp():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.CycLiqCPSP(osi, g0=1.0, kappa=1.0, h=1.0, big_m=1.0, dre1=1.0, dre2=1.0, rdr=1.0, alpha=1.0, dir=1.0, lambdac=1.0, ksi=1.0, e0=1.0, np=1.0, nd=1.0, ein=1.0, rho=1.0)




from o3seespy.command.element.base_element import ElementBase


class StdBrick(ElementBase):
    """
    The StdBrick Element Class
    
    This element is used to construct an eight-node brick element object, which uses a trilinear isoparametric
    formulation.

    
    """
    op_type = 'stdBrick'

    def __init__(self, osi, ele_nodes, mat, b1, b2, b3):
        """
        Initial method for StdBrick

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of eight element nodes in bottom and top faces and in counter-clockwise order
        mat: obj
            Object of ndmaterial
        b1: float
            Body forces in global x,y,z directions
        b2: float
            Body forces in global x,y,z directions
        b3: float
            Body forces in global x,y,z directions

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.StdBrick(osi, ele_nodes=ele_nodes, mat=mat, b1=1.0, b2=1.0, b3=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.b1 = float(b1)
        self.b2 = float(b2)
        self.b3 = float(b3)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.b1, self.b2, self.b3]
        self.to_process(osi)


class BbarBrick(ElementBase):
    """
    The BbarBrick Element Class
    
    This command is used to construct an eight-node mixed volume/pressure brick element object, which uses a trilinear
    isoparametric formulation.

    
    """
    op_type = 'bbarBrick'

    def __init__(self, osi, ele_nodes, mat, b1, b2, b3):
        """
        Initial method for BbarBrick

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of eight element nodes in bottom and top faces and in counter-clockwise order
        mat: obj
            Object of ndmaterial
        b1: float
            Body forces in global x,y,z directions
        b2: float
            Body forces in global x,y,z directions
        b3: float
            Body forces in global x,y,z directions

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.BbarBrick(osi, ele_nodes=ele_nodes, mat=mat, b1=1.0, b2=1.0, b3=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.b1 = float(b1)
        self.b2 = float(b2)
        self.b3 = float(b3)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.b1, self.b2, self.b3]
        self.to_process(osi)


class N20NodeBrick(ElementBase):
    """
    The N20NodeBrick Element Class
    
    The element is used to construct a twenty-node three dimensional element object

    
    """
    op_type = '20NodeBrick'

    def __init__(self, osi, ele_nodes, mat, bf1, bf2, bf3, mass_den):
        """
        Initial method for N20NodeBrick

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of twenty element nodes, input order is shown in notes below
        mat: obj
            Material object associated with previsouly-defined ndmaterial object
        bf1: float
            Body force in the direction of global coordinates x, y and z
        bf2: float
            Body force in the direction of global coordinates x, y and z
        bf3: float
            Body force in the direction of global coordinates x, y and z
        mass_den: float
            Mass density (mass/volume)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> coords = [[0, 0, 0], [-1, 0, 0], [-1, -1, 0], [0, -1, 0],
        >>>           [0, 0, -1], [-1, 0, -1], [-1, -1, -1], [0, -1, -1],
        >>>           [-0.5, 0, 0], [-1, -0.5, 0], [-0.5, -1, 0], [0, -0.5, 0],
        >>>           [-0.5, 0, -1], [-1, -0.5, -1], [-0.5, -1, -1], [0, -0.5, -1],
        >>>           [0, 0, -0.5], [-1, 0, -0.5], [-1, -1, -0.5], [0, -1, -0.5],
        >>>           ]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.N20NodeBrick(osi, ele_nodes=ele_nodes, mat=mat, bf1=1.0, bf2=1.0, bf3=1.0, mass_den=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.bf1 = float(bf1)
        self.bf2 = float(bf2)
        self.bf3 = float(bf3)
        self.mass_den = float(mass_den)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.bf1, self.bf2, self.bf3, self.mass_den]
        self.to_process(osi)


class SSPbrick(ElementBase):
    """
    The SSPbrick Element Class
    
    This command is used to construct a SSPbrick element object.

    
    """
    op_type = 'SSPbrick'

    def __init__(self, osi, ele_nodes, mat, b1, b2, b3):
        """
        Initial method for SSPbrick

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of eight element nodes in bottom and top faces and in counter-clockwise order
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        b1: float
            Constant body forces in global x-, y-, and z-directions, respectively (optional, default = 0.0)
        b2: float
            Constant body forces in global x-, y-, and z-directions, respectively (optional, default = 0.0)
        b3: float
            Constant body forces in global x-, y-, and z-directions, respectively (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.SSPbrick(osi, ele_nodes=ele_nodes, mat=mat, b1=1.0, b2=1.0, b3=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.b1 = float(b1)
        self.b2 = float(b2)
        self.b3 = float(b3)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.b1, self.b2, self.b3]
        self.to_process(osi)



import o3seespy as o3  # for testing only


def test_concrete01():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Concrete01(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0)


def test_concrete02():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Concrete02(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0, lamb=1.0, ft=1.0, ets=1.0)


def test_concrete04():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Concrete04(osi, fc=1.0, epsc=1.0, epscu=1.0, ec=1.0, fct=1.0, et=1.0, beta=1.0)


def test_concrete06():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Concrete06(osi, fc=1.0, e0=1.0, n=1.0, k=1.0, alpha1=1.0, fcr=1.0, ecr=1.0, b=1.0, alpha2=1.0)


def test_concrete07():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Concrete07(osi, fc=1.0, epsc=1.0, ec=1.0, ft=1.0, et=1.0, xp=1.0, xn=1.0, r=1.0)


def test_concrete01with_sitc():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Concrete01WithSITC(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0, end_strain_sitc=0.01)


def test_confined_concrete01():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ConfinedConcrete01(osi, sec_type='R', fpc=1.0, ec=1.0, epscu_type=1, epscu_val=1.0, nu=1, l1=1.0, l2=1.0, l3=1.0, phis=1.0, big_s=1.0, fyh=1.0, es0=1.0, ha_ratio=1.0, mu=1.0, phi_lon=1.0, gravel=1, tol=1.0, max_num_iter=1, epscu_limit=1.0, st_ratio=1)


def test_concrete_d():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ConcreteD(osi, fc=1.0, epsc=1.0, ft=1.0, epst=1.0, ec=1.0, alphac=1.0, alphat=1.0, cesp=0.25, etap=1.15)


def test_frp_confined_concrete():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.FRPConfinedConcrete(osi, fpc1=1.0, fpc2=1.0, epsc0=1.0, big_d=1.0, c=1.0, ej=1.0, sj=1.0, tj=1.0, eju=1.0, big_s=1.0, fyl=1.0, fyh=1.0, dlong=1.0, dtrans=1.0, es=1.0, nu0=1.0, k=1.0, use_buck=1.0)


def test_concrete_cm():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ConcreteCM(osi, fpcc=1.0, epcc=1.0, ec=1.0, rc=1.0, xcrn=1.0, ft=1.0, et=1.0, rt=1.0, xcrp=1.0, gap_close=0)





def mesh(osi, tag, args):
    """
    Create a mesh object. See below for available mesh types... toctree:::maxdepth:
    2linemeshtrimeshquadmeshtetmeshpartmeshbgmesh

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    tag: None
            
    args: None
            
    """
    _parameters = [tag, *args]
    return osi.to_process("mesh", _parameters)


def remesh(osi):
    r"""
    * :math:`\alpha \ge 0` for updating moving mesh.* :math:`\alpha < 0` for updating background mesh.If there are nodes
    shared by different mesh in the domain,the principles to decide what element should be used fora triangle:#. If all 3
    nodes share the same mesh, use that mesh for the triangle.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    """
    _parameters = []
    return osi.to_process("remesh", _parameters)


def integrator(osi):
    """
    Create a PFEM Integrator.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    """
    _parameters = []
    return osi.to_process("integrator", _parameters)


def system(osi, compressible=False, mumps=False):
    """
    Create a incompressible PFEM system of equations using the Umfpack solver

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    compressible: None
            
    mumps: None
            
    """
    _parameters = []
    if compressible:
        _parameters += ['-compressible']
    if mumps:
        _parameters += ['-mumps']
    return osi.to_process("system", _parameters)


def test(osi, tolv, tolp, tolrv, tolrp, tolrelv, tolrelp, max_iter, maxincr, p_flag=0, n_type=2):
    """
    Create a PFEM test, which check both increments and residual forvelocities and pressures.

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    tolv: float
            Tolerance for velocity increments
    tolp: float
            Tolerance for pressure increments
    tolrv: float
            Tolerance for relative velocity increments
    tolrp: float
            Tolerance for relative pressure increments
    tolrelv: None
            
    tolrelp: None
            
    max_iter: int
            Max number of iterations to check
    maxincr: int
            Max times for error increasing
    p_flag: int, optional
            Print flag : * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. * 2 print
            information on norms and number of iterations at end of successful test. * 4 at each step it will print the norms
            and also the :math:`\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter`` it
            will print an error message **but return a successfull test**.
    n_type: int, optional
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). 
    """
    tolv = float(tolv)
    tolp = float(tolp)
    tolrv = float(tolrv)
    tolrp = float(tolrp)
    max_iter = int(max_iter)
    maxincr = int(maxincr)
    p_flag = int(p_flag)
    n_type = int(n_type)
    _parameters = [tolv, tolp, tolrv, tolrp, tolrelv, tolrelp, max_iter, maxincr, p_flag, n_type]
    return osi.to_process("test", _parameters)


def analysis(osi, dtmax, dtmin, gravity, ratio=0.5):
    """
    Create a OpenSees PFEMAnalysis object. 

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    dtmax: float
            Maximum time steps.
    dtmin: float
            Mimimum time steps.
    gravity: float
            Gravity acceleration used to move isolated particles.
    ratio: float, optional
            The ratio to reduce time steps if it was not converged. 
    """
    dtmax = float(dtmax)
    dtmin = float(dtmin)
    gravity = float(gravity)
    ratio = float(ratio)
    _parameters = [dtmax, dtmin, gravity, ratio]
    return osi.to_process("analysis", _parameters)



import o3seespy as o3  # for testing only
import pytest


def test_elastic_beam_column2d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    transf = o3.geom_transf.Linear2D(osi, [])
    o3.element.ElasticBeamColumn2D(osi, ele_nodes=ele_nodes, area=1.0, e_mod=1.0, iz=1.0, transf=transf, mass=1.0)

def test_elastic_beam_column3d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    transf = o3.geom_transf.Linear2D(osi, [])
    o3.element.ElasticBeamColumn3D(osi, ele_nodes=ele_nodes, area=1.0, e_mod=1.0, g_mod=1.0, jxx=1.0, iy=1.0, iz=1.0, transf=transf, mass=1.0)


def test_mod_elastic_beam2dmass():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    transf = o3.geom_transf.Linear2D(osi, [])
    o3.element.ModElasticBeam2D(osi, ele_nodes=ele_nodes, area=1.0, e_mod=1.0, iz=1.0, k11=1.0, k33=1.0, k44=1.0, transf=transf, mass=1.0)


def test_elastic_timoshenko_beam2d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    transf = o3.geom_transf.Linear2D(osi, [])
    o3.element.ElasticTimoshenkoBeam2D(osi, ele_nodes=ele_nodes, e_mod=1.0, g_mod=1.0, area=1.0, iz=1.0, avy=1.0, transf=transf, mass=1.0)


def test_elastic_timoshenko_beam3d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    transf = o3.geom_transf.Linear2D(osi, [])
    o3.element.ElasticTimoshenkoBeam3D(osi, ele_nodes=ele_nodes, e_mod=1.0, g_mod=1.0, area=1.0, iz=1.0, jxx=1.0, iy=1.0, iz_2=1, avy=1.0, avz=1.0, transf=transf, mass=1.0, c_mass=True)


def test_disp_beam_column():
    osi = o3.OpenSeesInstance(ndm=2)
    i_node = o3.node.Node(osi, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0)
    transf = o3.geom_transf.Linear2D(osi, [])
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    integration = o3.beam_integration.Lobatto(osi, sec, 5)
    o3.element.DispBeamColumn(osi, ele_nodes=[i_node, j_node], transf=transf, integration=integration, c_mass=1, mass=0.0)


def test_force_beam_column():
    osi = o3.OpenSeesInstance(ndm=2)
    i_node = o3.node.Node(osi, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0)
    transf = o3.geom_transf.Linear2D(osi, [])
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    integration = o3.beam_integration.Lobatto(osi, sec, 5)
    o3.element.ForceBeamColumn(osi, ele_nodes=[i_node, j_node], transf=transf, integration=integration, max_iter=10, tol=1e-12, mass=0.0)


@pytest.mark.skip()  # can't find int_type that works with this!
def test_nonlinear_beam_column():
    osi = o3.OpenSeesInstance(ndm=2)
    i_node = o3.node.Node(osi, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0)
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    transf = o3.geom_transf.Linear2D(osi, [])
    o3.element.NonlinearBeamColumn(osi, ele_nodes=[i_node, j_node], num_intgr_pts=1, sec=sec, transf=transf,
                                   max_iter=10, tol=1e-12, mass=0.0, int_type="radau")


@pytest.mark.skip()
def test_disp_beam_column_int():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
    sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
    transf = o3.geom_transf.Linear2D(osi, [])
    o3.element.DispBeamColumnInt(osi, ele_nodes=ele_nodes, num_intgr_pts=4, sec=sec, transf=transf, c_rot=1.0, mass=1.0)


@pytest.mark.skip()
def test_mvlem():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    mat_conc = [o3.uniaxial_material.Concrete01(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0),
                o3.uniaxial_material.Concrete01(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0)]
    mat_steel = [o3.uniaxial_material.Steel02(osi, fy=1.0, e0=1.0, b=1.0, params=[15, 0.925, 0.15])]
    mat_shear = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    o3.element.MVLEM(osi, dens=1.0, ele_nodes=ele_nodes, m=1, c=1.0, thick=[1.0, 1.0], widths=[1, 1], rho=[1., 1.],
                     mat_concretes=mat_conc, mat_steels=mat_steel, mat_shear=mat_shear)


@pytest.mark.skip()
def test_sfimvlem():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    mats = [o3.uniaxial_material.Elastic(osi, 1.0, 1.0), o3.uniaxial_material.Elastic(osi, 1.0, 1.0)]
    mat_tags = [x.tag for x in mats]  # TODO: should pass in mats not mat tags
    o3.element.SFIMVLEM(osi, ele_nodes=ele_nodes, m=1, c=1.0, thick=[1.0, 1.0], widths=[1.0, 1.0], mat_tags=mat_tags)




import o3seespy as o3  # for testing only


def test_plain():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.constraints.Plain(osi)


def test_lagrange():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.constraints.Lagrange(osi, alpha_m=1.0)


def test_penalty():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.constraints.Penalty(osi, alpha_m=1.0)


def test_transformation():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.constraints.Transformation(osi)




import o3seespy as o3  # for testing only


def test_py_simple1():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.PySimple1(osi, soil_type=1, pult=1.0, y50=1.0, cd=1.0, c=0.0)


def test_tz_simple1():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.TzSimple1(osi, soil_type=1, tult=1.0, z50=1.0, c=0.0)


def test_qz_simple1():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.QzSimple1(osi, qz_type=1, qult=1.0, z50=1.0, suction=0.0, c=0.0)


def test_py_liq1():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.PyLiq1(osi, soil_type=1, pult=1.0, y50=1.0, cd=1.0, c=1.0, p_res=1.0, ele1=1.0, ele2=1.0)


def test_tz_liq1():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.TzLiq1(osi, tz_type=1, tult=1.0, z50=1.0, c=1.0, ele1=1.0, ele2=1.0)




import o3seespy as o3  # for testing only
import pytest


def test_steel02():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Steel02(osi, fy=1.0, e0=1.0, b=1.0, params=[15, 0.925, 0.15])



def test_hysteretic():
    osi = o3.OpenSeesInstance(ndm=2)
    p1 = [0.5, 0.5]
    p2 = [1.0, 1.0]
    p3 = [0, 1.5]
    n1 = [-0.5, -0.5]
    n2 = [-1.0, -1.0]
    n3 = [0, -1.5]
    o3.uniaxial_material.Hysteretic(osi, p1=p1, p2=p2, p3=p3, n1=n1, n2=n2, n3=n3, pinch_x=1, pinch_y=0, damage1=0, damage2=0)


def test_reinforcing_steel_ga_buck():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ReinforcingSteelGABuck(osi, fy=1.0, fu=1.0, es=1.0, esh=1.0, eps_sh=1.0, eps_ult=1.0, lsr=1.0, beta=1.0, r=1.0, gamma=1.0)


def test_reinforcing_steel_dm_buck():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ReinforcingSteelDMBuck(osi, fy=1.0, fu=1.0, es=1.0, esh=1.0, eps_sh=1.0, eps_ult=1.0, lsr_2=1, alpha=1.0)


def test_reinforcing_steel_cm_fatigue():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ReinforcingSteelCMFatigue(osi, fy=1.0, fu=1.0, es=1.0, esh=1.0, eps_sh=1.0, eps_ult=1.0, cf=1.0, alpha_2=1, cd=1.0)


def test_reinforcing_steel_iso_hard():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ReinforcingSteelIsoHard(osi, fy=1.0, fu=1.0, es=1.0, esh=1.0, eps_sh=1.0, eps_ult=1.0, a1=4.3, limit=1.0)


def test_reinforcing_steel_mp_curve_params():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.ReinforcingSteelMPCurveParams(osi, fy=1.0, fu=1.0, es=1.0, esh=1.0, eps_sh=1.0, eps_ult=1.0, r1=0.333, r2=18.0, r3=4.0)


@pytest.mark.skip()  # not connected in openseespy - works in tcl
def test_dodd_restrepo():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.DoddRestrepo(osi, fy=1.0, fsu=1.0, esh=1.0, esu=1.0, youngs=1.0, eshi=1.0, fshi=1.0, omega_fac=1.0)


def test_ramberg_osgood_steel():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.RambergOsgoodSteel(osi, fy=1.0, e0=1.0, a=1.0, n=1.0)


def test_steel_mpf():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.SteelMPF(osi, fyp=1.0, fyn=1.0, e0=1.0, bp=1.0, bn=1.0, params=[1.0, 1.0, 1.0], a1=0.0, a2=1.0, a3=0.0, a4=1.0)


def test_steel01thermal():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.uniaxial_material.Steel01Thermal(osi, fy=1.0, e0=1.0, b=1.0, a1=1.0, a2=1.0, a3=1.0, a4=1.0)




import o3seespy as o3  # for testing only


def test_load_control():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.LoadControl(osi, incr=1.0, num_iter=1, min_incr=None, max_incr=None)


def test_displacement_control():
    osi = o3.OpenSeesInstance(ndm=2)
    node = o3.node.Node(osi, 0.0, 0.0)
    o3.integrator.DisplacementControl(osi, node, dof=1, incr=1.0, num_iter=1, d_umin=None, d_umax=None)


def test_parallel_displacement_control():
    osi = o3.OpenSeesInstance(ndm=2)
    node = o3.node.Node(osi, 0.0, 0.0)
    o3.integrator.ParallelDisplacementControl(osi, node, dof=1, incr=1.0, num_iter=1, d_umin=None, d_umax=None)


def test_min_unbal_disp_norm():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.MinUnbalDispNorm(osi, dlambda1=1.0, jd=1, min_lambda=None, max_lambda=None, det=None)


def test_arc_length():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.ArcLength(osi, s=1.0, alpha=1.0)


def test_central_difference():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.CentralDifference(osi)


def test_newmark():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.Newmark(osi, gamma=1.0, beta=1.0, form=1)


def test_hht():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.HHT(osi, alpha=1.0, gamma=1.0, beta=1.0)


def test_generalized_alpha():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.GeneralizedAlpha(osi, alpha_m=1.0, alpha_f=1.0, gamma=1.0, beta=1.0)


def test_trbdf2():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.TRBDF2(osi)


def test_explicit_difference():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.integrator.ExplicitDifference(osi)


import o3seespy as o3
import numpy as np


def test_build_regular_node_mesh_2dxy():
    xs = [1, 2, 3]
    ys = [3, 4]
    zs = 4.5
    osi = o3.OpenSeesInstance(ndm=3, ndf=3)
    sn = o3.node.build_regular_node_mesh(osi, xs, ys, zs)
    sn = np.array(sn)
    print(sn.shape)
    assert len(sn.shape) == 2  # two axes
    assert sn.shape[0] == 3  # y-axis
    assert sn.shape[1] == 2  # x-axis
    assert sn[2][1].x == 3.0
    assert sn[2][1].y == 4.0
    assert sn[2][1].z == 4.5, sn[2][1].z
    # No z-axis
    zs = None
    osi = o3.OpenSeesInstance(ndm=2, ndf=3)
    sn = o3.node.build_regular_node_mesh(osi, xs, ys, zs)
    sn = np.array(sn)
    assert len(sn.shape) == 2  # two axes

    assert sn.shape[0] == 3  # x-axis
    assert sn.shape[1] == 2  # y-axis
    assert sn[2][1].x == 3.0
    assert sn[2][1].y == 4.0
    assert not hasattr(sn[2][1], 'z'), sn[1][2].z


def test_build_regular_node_mesh_2dxz():
    xs = [1, 2, 3]
    ys = [3]
    zs = [4.5, 6.0]

    osi = o3.OpenSeesInstance(ndm=3, ndf=3)
    sn = o3.node.build_regular_node_mesh(osi, xs, ys, zs)

    sn = np.array(sn)
    assert len(sn.shape) == 3  # three axes
    assert sn.shape[0] == 3  # y-axis
    assert sn.shape[1] == 1  # x-axis
    assert sn.shape[2] == 2  # x-axis
    assert sn[2][0][1].x == 3.0




from o3seespy.command.element.base_element import ElementBase


class TwoNodeLink(ElementBase):
    """
    The TwoNodeLink Element Class
    
    This command is used to construct a twoNodeLink element object, which is defined by two nodes. The element can have
    zero or non-zero length. This element can have 1 to 6 degrees of freedom, where only the transverse and rotational
    degrees of freedom are coupled as long as the element has non-zero length. In addition, if the element length is
    larger than zero, the user can optionally specify how the P-Delta moments around the local x- and y-axis are
    distributed among a moment at node i, a moment at node j, and a shear couple. The sum of these three ratios
    is always equal to 1. In addition the shear center can be specified as a fraction of the element length
    from the iNode. The element does not contribute to the Rayleigh damping by default. If the element has
    non-zero length, the local x-axis is determined from the nodal geometry unless the optional x-axis
    vector is specified in which case the nodal geometry is ignored and the user-defined orientation
    is utilized. It is important to recognize that if this element has zero length, it does not
    consider the geometry as given by the nodal coordinates, but utilizes the user-defined
    orientation vectors to determine the directions of the springs.

    
    """
    op_type = 'twoNodeLink'

    def __init__(self, osi, ele_nodes, mats: list=None, dir=None, p_delta_vals: list=None, shear_dist=None, do_rayleigh=False, orient: list=None, mass: float=None):
        """
        Initial method for TwoNodeLink

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        mats: list, optional
            A list of objects associated with previously-defined uniaxial_material objects
        dir: None, optional
            
        p_delta_vals: list, optional
            P-delta moment contribution ratios, size of ratio vector is 2 for 2d-case and 4 for 3d-case (entries:
            ``[my_inode, my_jnode, mz_inode, mz_jnode]``) ``my_inode`` + ``my_jnode`` <= 1.0, ``mz_inode`` + ``mz_jnode`` <=
            1.0. remaining p-delta moments are resisted by shear couples. 
        shear_dist: None, optional
            
        do_rayleigh: bool
            To include rayleigh damping from the element (optional, default = no rayleigh damping contribution)
        orient: list, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> mats = [o3.uniaxial_material.Elastic(osi, 1.0),
        >>>         o3.uniaxial_material.Elastic(osi, 1.0)]
        >>> p_delta_vals = [1.0, 1.0]
        >>> o3.element.TwoNodeLink(osi, ele_nodes=ele_nodes, mats=mats, dir=[1, 1], p_delta_vals=p_delta_vals)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        if mats is None:
            self.mats = None
        else:
            self.mats = [x.tag for x in mats]
        self.dir = dir
        self.p_delta_vals = p_delta_vals
        self.shear_dist = shear_dist
        self.do_rayleigh = do_rayleigh
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags]
        if getattr(self, 'mats') is not None:
            self._parameters += ['-mat', *self.mats]
        if getattr(self, 'dir') is not None:
            self._parameters += ['-dir', *self.dir]
        if getattr(self, 'p_delta_vals') is not None:
            self._parameters += ['-pDelta', *self.p_delta_vals]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', *self.shear_dist]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'orient') is not None:
            self._parameters += ['--orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)



import o3seespy as o3  # for testing only


def test_linear2d():
    osi = o3.OpenSeesInstance(ndm=2)
    d_i = [1.0, 1.0]
    d_j = [1.0, 1.0]
    o3.geom_transf.Linear2D(osi, d_i=d_i, d_j=d_j)

def test_linear3d():
    osi = o3.OpenSeesInstance(ndm=2)
    vecxz = [1.0, 1.0]
    d_i = [1.0, 1.0]
    d_j = [1.0, 1.0]
    o3.geom_transf.Linear3D(osi, vecxz=vecxz, d_i=d_i, d_j=d_j)


def test_p_delta2d():
    osi = o3.OpenSeesInstance(ndm=2)
    d_i = [1.0, 1.0]
    d_j = [1.0, 1.0]
    o3.geom_transf.PDelta2D(osi, d_i=d_i, d_j=d_j)

def test_p_delta3d():
    osi = o3.OpenSeesInstance(ndm=2)
    vecxz = [1.0, 1.0]
    d_i = [1.0, 1.0]
    d_j = [1.0, 1.0]
    o3.geom_transf.PDelta3D(osi, vecxz=vecxz, d_i=d_i, d_j=d_j)


def test_corotational2d():
    osi = o3.OpenSeesInstance(ndm=2)
    d_i = [1.0, 1.0]
    d_j = [1.0, 1.0]
    o3.geom_transf.Corotational2D(osi, d_i=d_i, d_j=d_j)

def test_corotational3d():
    osi = o3.OpenSeesInstance(ndm=2)
    vecxz = [1.0, 1.0]
    o3.geom_transf.Corotational3D(osi, vecxz=vecxz)




import o3seespy as o3  # for testing only
import pytest


def test_simple_contact2d():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    i_node = o3.node.Node(osi, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0)
    c_node = o3.node.Node(osi, 0.0, 0.0)
    l_node = o3.node.Node(osi, 0.0, 1.0)
    mat = o3.nd_material.ContactMaterial2D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
    o3.element.SimpleContact2D(osi, i_node=i_node, j_node=j_node, c_node=c_node, l_node=l_node, mat=mat,
                               g_tol=1.0, f_tol=1.0)


@pytest.mark.skip()
def test_simple_contact3d():
    osi = o3.OpenSeesInstance(ndm=3, ndf=3)
    i_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    k_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
    c_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
    l_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    lagr_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    mat = o3.nd_material.ContactMaterial3D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
    o3.element.SimpleContact3D(osi, i_node=i_node, j_node=j_node, k_node=k_node, l_node=l_node, c_node=c_node,
                               lagr_node=lagr_node, mat=mat, g_tol=1.0, f_tol=1.0)


def test_beam_contact2d():
    osi = o3.OpenSeesInstance(ndm=2)
    i_node = o3.node.Node(osi, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0)
    s_node = o3.node.Node(osi, 0.0, 1.0)
    l_node = o3.node.Node(osi, 0.0, 1.0)
    mat = o3.nd_material.ContactMaterial2D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
    o3.element.BeamContact2D(osi, i_node=i_node, j_node=j_node, s_node=s_node, l_node=l_node, mat=mat, width=1.0,
                             g_tol=1.0, f_tol=1.0, c_flag=1)


@pytest.mark.skip()
def test_beam_contact3d():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    i_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    osi.reset_model_params(ndm=3, ndf=3)
    c_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    l_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    osi.reset_model_params(ndm=3, ndf=6)
    mat = o3.nd_material.ContactMaterial3D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
    crd_transf = o3.geom_transf.Linear3D(osi, vecxz=[1.0, 1.0], d_i=[1.0, 1.0], d_j=[1.0, 1.0])
    o3.element.BeamContact3D(osi, i_node=i_node, j_node=j_node, c_node=c_node, l_node=l_node, radius=1.0,
                             crd_transf=crd_transf, mat=mat, g_tol=1.0, f_tol=1.0, c_flag=1)


def test_beam_end_contact3d():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    i_node = o3.node.Node(osi, 0.0, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    osi.reset_model_params(ndm=3, ndf=3)
    c_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    l_node = o3.node.Node(osi, 0.0, 1.0, 0.0)
    o3.element.BeamEndContact3D(osi, i_node=i_node, j_node=j_node, c_node=c_node, l_node=l_node, radius=1.0,
                                g_tol=1.0, f_tol=1.0, c_flag=1.0)




import o3seespy as o3  # for testing only
import pytest


def test_band_gen():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.system.BandGen(osi)


def test_band_spd():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.system.BandSPD(osi)


def test_profile_spd():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.system.ProfileSPD(osi)


def test_super_lu():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.system.SuperLU(osi)


def test_umf_pack():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.system.UmfPack(osi)


def test_full_general():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.system.FullGeneral(osi)


@pytest.mark.skip()
def test_sparse_sym():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.system.SparseSYM(osi)


#
# def test_mumps():
#     osi = o3.OpenSeesInstance(ndm=2)
#     o3.system.Mumps(osi, icntl14=20.0, icntl7=7)




from o3seespy.base_model import OpenSeesObject


class TestBase(OpenSeesObject):
    op_base_type = "test"


class EnergyIncr(TestBase):
    op_type = "EnergyIncr"

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)  # changed to avoid python function iter
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class FixedNumIter(TestBase):
    op_type = "FixedNumIter"

    def __init__(self, osi, max_iter, p_flag=0, n_type=2):
        self.osi = osi
        self.max_iter = int(max_iter)  # changed to avoid python function iter
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class NormDispIncr(TestBase):
    op_type = "NormDispIncr"

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)  # changed to avoid python function iter
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class RelativeNormDispIncr(TestBase):
    op_type = "RelativeNormDispIncr"

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        """
        Initial method for RelativeNormDispIncr

        Parameters
        ----------
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int
            Print flag (optional):
            * 0 print nothing.
            * 1 print information on norms each time ``test()`` is invoked.
            * 2 print information on norms and number of iterations at end of successful test.
            * 4 at each step it will print the norms and also the :math:`\\delta u` and :math:`R(u)` vectors.
            * 5 if it fails to converge at end of ``numiter``
                it will print an error message **but return a successfull test**.
        n_type: int
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). (optional)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)  # changed to avoid python function iter
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        self.to_process(osi)


class NormUnbalance(TestBase):
    """
    The NormUnbalance Test Class

    Create a NormUnbalance test, which uses the norm of the right hand side of the matrix equation to determine if
    convergence has been reached.
    """
    op_type = 'NormUnbalance'

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2, max_incr: int = None):
        """
        Initial method for NormUnbalance

        Parameters
        ----------
        tol: float
            Tolerance criteria used to check for convergence.
        max_iter: int
            Max number of iterations to check
        p_flag: int
            Print flag (optional): * 0 print nothing. * 1 print information on norms each time ``test()`` is invoked. *
            2 print information on norms and number of iterations at end of successful test. * 4 at each step it will print the
            norms and also the :math:`\\delta u` and :math:`r(u)` vectors. * 5 if it fails to converge at end of ``numiter``
            it will print an error message **but return a successfull test**.
        n_type: int
            Type of norm, (0 = max-norm, 1 = 1-norm, 2 = 2-norm). (optional)
        max_incr: int (default=True)
            Maximum times of error increasing. (optional)
        """
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        if max_incr is None:
            self.max_incr = None
        else:
            self.max_incr = int(max_incr)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type]
        special_pms = ['max_incr']
        packets = [False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class RelativeNormUnbalance(TestBase):
    op_type = "RelativeNormUnbalance"

    def __init__(self, osi, tol, max_iter, p_flag=0, n_type=2):
        self.osi = osi
        self.tol = float(tol)
        self.max_iter = int(max_iter)  # changed to avoid python function iter
        self.p_flag = int(p_flag)
        self.n_type = int(n_type)
        self._parameters = [self.op_type, self.tol, self.max_iter, self.p_flag, self.n_type, self.n_type]
        self.to_process(osi)
